from src.acrod.jacobian import Jacobian

import numpy

def test_planar_3Rs():
    A = numpy.array(
        [[9, 1, 0, 0],
         [1, 9, 1, 0],
         [0, 1, 9, 1],
         [0, 0, 1, 9]]
    )
    jac = Jacobian(A, robot_type = 'planar')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[-a_y + r_{(1,2)y}, -a_y + r_{(2,3)y}, -a_y + r_{(3,4)y}],\n[ a_x - r_{(1,2)x},  a_x - r_{(2,3)x},  a_x - r_{(3,4)x}],\n[                1,                 1,                 1]]), None, None, None]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(2)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[0.0, -0.17030618, -0.06929525], [0.0, -0.05394987, 0.1251587], [1.0, 1.0, 1.0]])).all()

def test_planar_3RPRp():
    A = numpy.array(
        [[9, 1, 0, 1, 0, 1, 0, 0],
         [0, 9, 2, 0, 0, 0, 0, 0],
         [0, 1, 9, 0, 0, 0, 0, 1],
         [0, 0, 0, 9, 2, 0, 0, 0],
         [0, 0, 0, 1, 9, 0, 0, 1],
         [0, 0, 0, 0, 0, 9, 2, 0],
         [0, 0, 0, 0, 0, 1, 9, 1],
         [0, 0, 0, 0, 0, 0, 0, 9]]
        )
    jac = Jacobian(A, robot_type = 'planar')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[cos(\\phi_{(2,3)}), 0, 0],\n[sin(\\phi_{(2,3)}), 0, 0],\n[                0, 0, 0]]), Matrix([\n[-a_y + r_{(1,2)y}, 0, 0, -a_y + r_{(3,8)y}, 0, 0],\n[ a_x - r_{(1,2)x}, 0, 0,  a_x - r_{(3,8)x}, 0, 0],\n[                1, 0, 0,                 1, 0, 0]]), Matrix([\n[-cos(\\phi_{(2,3)}), cos(\\phi_{(4,5)}),                 0],\n[-sin(\\phi_{(2,3)}), sin(\\phi_{(4,5)}),                 0],\n[-cos(\\phi_{(2,3)}),                 0, cos(\\phi_{(6,7)})],\n[-sin(\\phi_{(2,3)}),                 0, sin(\\phi_{(6,7)})],\n[                 0,                 0,                 0],\n[                 0,                 0,                 0]]), Matrix([\n[ a_y - r_{(1,2)y}, -a_y + r_{(1,4)y},                 0,  a_y - r_{(3,8)y}, -a_y + r_{(5,8)y},                 0],\n[-a_x + r_{(1,2)x},  a_x - r_{(1,4)x},                 0, -a_x + r_{(3,8)x},  a_x - r_{(5,8)x},                 0],\n[ a_y - r_{(1,2)y},                 0, -a_y + r_{(1,6)y},  a_y - r_{(3,8)y},                 0, -a_y + r_{(7,8)y}],\n[-a_x + r_{(1,2)x},                 0,  a_x - r_{(1,6)x}, -a_x + r_{(3,8)x},                 0,  a_x - r_{(7,8)x}],\n[               -1,                 1,                 0,                -1,                 1,                 0],\n[               -1,                 0,                 1,                -1,                 0,                 1]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(2)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[1.5116862461994787, -1.001002195991253, 0.19757451735055986], [-0.41256134092451774, 1.381649776595281, -0.27270548341604917], [-9.331363937770648, 12.149247344303607, -1.3312812081939858]])).all()

    
def test_planar_5Rp():
    A = numpy.array(
        [[9, 1, 0, 1, 0],
         [1, 9, 1, 0, 0],
         [0, 0, 9, 0, 1],
         [1, 0, 0, 9, 1],
         [0, 0, 0, 0, 9]]
    )
    jac = Jacobian(A, robot_type = 'planar')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[0, -a_y + r_{(1,4)y}],\n[0,  a_x - r_{(1,4)x}],\n[0,                 1]]), Matrix([\n[0, 0, -a_y + r_{(4,5)y}],\n[0, 0,  a_x - r_{(4,5)x}],\n[0, 0,                 1]]), Matrix([\n[-a_y + r_{(1,2)y},  a_y - r_{(1,4)y}],\n[ a_x - r_{(1,2)x}, -a_x + r_{(1,4)x}],\n[                1,                -1]]), Matrix([\n[-a_y + r_{(2,3)y}, -a_y + r_{(3,5)y},  a_y - r_{(4,5)y}],\n[ a_x - r_{(2,3)x},  a_x - r_{(3,5)x}, -a_x + r_{(4,5)x}],\n[                1,                 1,                -1]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(2)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[-0.07248152798410853, -0.059799857441136486], [-0.09063783901508533, 0.0842378131702819], [0.2184837927966302, 0.6668966302219953]])).all()


def test_planar_4R4Psp():
    A = numpy.array(
        [[9, 2, 2, 0, 0, 0, 0],
         [1, 9, 0, 2, 0, 0, 0],
         [1, 0, 9, 2, 0, 0, 0],
         [0, 0, 0, 9, 1, 1, 0],
         [0, 0, 0, 1, 9, 0, 1],
         [0, 0, 0, 0, 0, 9, 1],
         [0, 0, 0, 0, 0, 0, 9]]
    )
    jac = Jacobian(A, robot_type = 'planar')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[cos(\\phi_{(1,2)}), 0, -a_y + r_{(4,5)y}],\n[sin(\\phi_{(1,2)}), 0,  a_x - r_{(4,5)x}],\n[                0, 0,                 1]]), Matrix([\n[cos(\\phi_{(2,4)}), 0, 0, -a_y + r_{(5,7)y}, 0],\n[sin(\\phi_{(2,4)}), 0, 0,  a_x - r_{(5,7)x}, 0],\n[                0, 0, 0,                 1, 0]]), Matrix([\n[                 0,                 0,  a_y - r_{(4,5)y}],\n[                 0,                 0, -a_x + r_{(4,5)x}],\n[-cos(\\phi_{(1,2)}), cos(\\phi_{(1,3)}),                 0],\n[-sin(\\phi_{(1,2)}), sin(\\phi_{(1,3)}),                 0],\n[                 0,                 0,                -1]]), Matrix([\n[                 0,                 0, -a_y + r_{(4,6)y},  a_y - r_{(5,7)y}, -a_y + r_{(6,7)y}],\n[                 0,                 0,  a_x - r_{(4,6)x}, -a_x + r_{(5,7)x},  a_x - r_{(6,7)x}],\n[-cos(\\phi_{(2,4)}), cos(\\phi_{(3,4)}),                 0,                 0,                 0],\n[-sin(\\phi_{(2,4)}), sin(\\phi_{(3,4)}),                 0,                 0,                 0],\n[                 0,                 0,                 1,                -1,                 1]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(2)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[0.7971752173945887, 2.4135074176540847, -2.878018570595505], [0.4831574969754604, 1.660978841059345, -3.3871377239062928], [0.0, 0.0, 9.466440213737489]])).all()


def test_spatial_6Rs():
    A = numpy.array(
        [[9, 1, 0, 0, 0, 0, 0],
         [1, 9, 1, 0, 0, 0, 0],
         [0, 1, 9, 1, 0, 0, 0],
         [0, 0, 1, 9, 1, 0, 0],
         [0, 0, 0, 1, 9, 1, 0],
         [0, 0, 0, 0, 1, 9, 1],
         [0, 0, 0, 0, 0, 1, 9]]
    )
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[                  -(a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                   -(a_y - r_{(2,3)y})*cos(\\beta_{(2,3)}) + (a_z - r_{(2,3)z})*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                   -(a_y - r_{(3,4)y})*cos(\\beta_{(3,4)}) + (a_z - r_{(3,4)z})*sin(\\beta_{(3,4)})*sin(\\phi_{(3,4)}),                   -(a_y - r_{(4,5)y})*cos(\\beta_{(4,5)}) + (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                   -(a_y - r_{(5,6)y})*cos(\\beta_{(5,6)}) + (a_z - r_{(5,6)z})*sin(\\beta_{(5,6)})*sin(\\phi_{(5,6)}),                   -(a_y - r_{(6,7)y})*cos(\\beta_{(6,7)}) + (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)})],\n[                   (a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                    (a_x - r_{(2,3)x})*cos(\\beta_{(2,3)}) - (a_z - r_{(2,3)z})*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                    (a_x - r_{(3,4)x})*cos(\\beta_{(3,4)}) - (a_z - r_{(3,4)z})*sin(\\beta_{(3,4)})*cos(\\phi_{(3,4)}),                    (a_x - r_{(4,5)x})*cos(\\beta_{(4,5)}) - (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                    (a_x - r_{(5,6)x})*cos(\\beta_{(5,6)}) - (a_z - r_{(5,6)z})*sin(\\beta_{(5,6)})*cos(\\phi_{(5,6)}),                    (a_x - r_{(6,7)x})*cos(\\beta_{(6,7)}) - (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)})],\n[-(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -(a_x - r_{(2,3)x})*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}) + (a_y - r_{(2,3)y})*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}), -(a_x - r_{(3,4)x})*sin(\\beta_{(3,4)})*sin(\\phi_{(3,4)}) + (a_y - r_{(3,4)y})*sin(\\beta_{(3,4)})*cos(\\phi_{(3,4)}), -(a_x - r_{(4,5)x})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}) + (a_y - r_{(4,5)y})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}), -(a_x - r_{(5,6)x})*sin(\\beta_{(5,6)})*sin(\\phi_{(5,6)}) + (a_y - r_{(5,6)y})*sin(\\beta_{(5,6)})*cos(\\phi_{(5,6)}), -(a_x - r_{(6,7)x})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}) + (a_y - r_{(6,7)y})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)})],\n[                                                                              sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                               sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                                                                               sin(\\beta_{(3,4)})*cos(\\phi_{(3,4)}),                                                                               sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                                                                               sin(\\beta_{(5,6)})*cos(\\phi_{(5,6)}),                                                                               sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)})],\n[                                                                              sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                               sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                                                                               sin(\\beta_{(3,4)})*sin(\\phi_{(3,4)}),                                                                               sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                                                                               sin(\\beta_{(5,6)})*sin(\\phi_{(5,6)}),                                                                               sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)})],\n[                                                                                                cos(\\beta_{(1,2)}),                                                                                                 cos(\\beta_{(2,3)}),                                                                                                 cos(\\beta_{(3,4)}),                                                                                                 cos(\\beta_{(4,5)}),                                                                                                 cos(\\beta_{(5,6)}),                                                                                                 cos(\\beta_{(6,7)})]]), None, None, None]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[0.0, -0.24717605269353116, -0.110060798347842, -0.6182914851868222, 0.07139447103785317, -0.5643889221863093], [0.0, 0.16473731751223517, -0.1737464627485653, 0.43285794014178935, -0.40411481397676463, 0.07700929758671417], [0.0, 0.2412633703752826, 0.16224527698950988, 0.06686037502989847, -0.022985669776265615, 0.27919046895597954], [0.4724950103528386, 0.7616470347417783, 0.7969646718055167, 0.45264097702195893, 0.6987395918153502, 0.4562381192065277], [0.21307795514111652, 0.30725509823719366, 0.05670864715460019, 0.5365398276183602, 0.08303055263692735, 0.20082922153309404], [0.855187844993456, 0.5705156431477658, 0.6013579975620347, 0.7122086487117851, 0.7105412797005703, 0.866899303472613]])).all()


def test_spatial_RSSRp():
    A = numpy.array(
        [[9, 1, 0, 1],
         [1, 9, 4, 0],
         [0, 0, 9, 4],
         [0, 0, 0, 9]]
    )
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[0],\n[0],\n[0],\n[0],\n[0],\n[0]]), Matrix([\n[                  -(a_y - r_{(1,4)y})*cos(\\beta_{(1,4)}) + (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}), 0, 0, 0, 0, 0, 0],\n[                   (a_x - r_{(1,4)x})*cos(\\beta_{(1,4)}) - (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}), 0, 0, 0, 0, 0, 0],\n[-(a_x - r_{(1,4)x})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}) + (a_y - r_{(1,4)y})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}), 0, 0, 0, 0, 0, 0],\n[                                                                              sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}), 0, 0, 0, 0, 0, 0],\n[                                                                              sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}), 0, 0, 0, 0, 0, 0],\n[                                                                                                cos(\\beta_{(1,4)}), 0, 0, 0, 0, 0, 0]]), Matrix([\n[                                                                              -(a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)})],\n[                                                                               (a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)})],\n[                                                            -(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)})],\n[                                                                                                                                          sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)})],\n[                                                                                                                                          sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)})],\n[                                                                                                                                                            cos(\\beta_{(1,2)})],\n[(r_{(2,3)x} - r_{(3,4)x})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}) + (r_{(2,3)y} - r_{(3,4)y})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (r_{(2,3)z} - r_{(3,4)z})*cos(\\beta_{(1,2)})]]), Matrix([\n[                  (a_y - r_{(1,4)y})*cos(\\beta_{(1,4)}) - (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                       0,        a_z - r_{(2,3)z},       -a_y + r_{(2,3)y},                 0,  a_z - r_{(3,4)z}, -a_y + r_{(3,4)y}],\n[                 -(a_x - r_{(1,4)x})*cos(\\beta_{(1,4)}) + (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),       -a_z + r_{(2,3)z},                       0,        a_x - r_{(2,3)x}, -a_z + r_{(3,4)z},                 0,  a_x - r_{(3,4)x}],\n[(a_x - r_{(1,4)x})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}) - (a_y - r_{(1,4)y})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),        a_y - r_{(2,3)y},       -a_x + r_{(2,3)x},                       0,  a_y - r_{(3,4)y}, -a_x + r_{(3,4)x},                 0],\n[                                                                            -sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                       1,                       0,                       0,                 1,                 0,                 0],\n[                                                                            -sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                       0,                       1,                       0,                 0,                 1,                 0],\n[                                                                                              -cos(\\beta_{(1,4)}),                       0,                       0,                       1,                 0,                 0,                 1],\n[                                                                                                                0, r_{(2,3)x} - r_{(3,4)x}, r_{(2,3)y} - r_{(3,4)y}, r_{(2,3)z} - r_{(3,4)z},                 0,                 0,                 0]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[-0.02032654749573748], [0.013547189835920806], [0.01984031747199766], [0.0626341203282414], [0.02526715383455612], [0.04691641116173456]])).all()


def test_spatial_RSSRSSR():
    A = numpy.array(
        [[9, 1, 1, 0, 1, 0],
         [1, 9, 0, 0, 0, 4],
         [1, 0, 9, 4, 0, 0],
         [0, 0, 0, 9, 0, 4],
         [0, 0, 0, 0, 9, 4],
         [0, 0, 0, 0, 0, 9]]
    )
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[                  -(a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), 0],\n[                   (a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0],\n[-(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0],\n[                                                                              sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0],\n[                                                                              sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), 0],\n[                                                                                                cos(\\beta_{(1,2)}), 0]]), Matrix([\n[0,                 0,  a_z - r_{(2,6)z}, -a_y + r_{(2,6)y}, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[0, -a_z + r_{(2,6)z},                 0,  a_x - r_{(2,6)x}, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[0,  a_y - r_{(2,6)y}, -a_x + r_{(2,6)x},                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[0,                 1,                 0,                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[0,                 0,                 1,                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[0,                 0,                 0,                 1, 0, 0, 0, 0, 0, 0, 0, 0, 0]]), Matrix([\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                                                                              0],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                                                                              0],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                                                                              0],\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                               -(a_y - r_{(1,3)y})*cos(\\beta_{(1,3)}) + (a_z - r_{(1,3)z})*sin(\\beta_{(1,3)})*sin(\\phi_{(1,3)})],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                (a_x - r_{(1,3)x})*cos(\\beta_{(1,3)}) - (a_z - r_{(1,3)z})*sin(\\beta_{(1,3)})*cos(\\phi_{(1,3)})],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                             -(a_x - r_{(1,3)x})*sin(\\beta_{(1,3)})*sin(\\phi_{(1,3)}) + (a_y - r_{(1,3)y})*sin(\\beta_{(1,3)})*cos(\\phi_{(1,3)})],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                                                                              0],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                                                                              0],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                                                                                              0],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                                           sin(\\beta_{(1,3)})*cos(\\phi_{(1,3)})],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                                           sin(\\beta_{(1,3)})*sin(\\phi_{(1,3)})],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                                                                             cos(\\beta_{(1,3)})],\n[                                                                                                                0, (r_{(3,4)x} - r_{(4,6)x})*sin(\\beta_{(1,3)})*cos(\\phi_{(1,3)}) + (r_{(3,4)y} - r_{(4,6)y})*sin(\\beta_{(1,3)})*sin(\\phi_{(1,3)}) + (r_{(3,4)z} - r_{(4,6)z})*cos(\\beta_{(1,3)})]]), Matrix([\n[                  -(a_y - r_{(1,5)y})*cos(\\beta_{(1,5)}) + (a_z - r_{(1,5)z})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}),                 0, -a_z + r_{(2,6)z},  a_y - r_{(2,6)y},                       0,                       0,                       0,                 0,                 0,                 0,                 0,  a_z - r_{(5,6)z}, -a_y + r_{(5,6)y}],\n[                   (a_x - r_{(1,5)x})*cos(\\beta_{(1,5)}) - (a_z - r_{(1,5)z})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}),  a_z - r_{(2,6)z},                 0, -a_x + r_{(2,6)x},                       0,                       0,                       0,                 0,                 0,                 0, -a_z + r_{(5,6)z},                 0,  a_x - r_{(5,6)x}],\n[-(a_x - r_{(1,5)x})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}) + (a_y - r_{(1,5)y})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}), -a_y + r_{(2,6)y},  a_x - r_{(2,6)x},                 0,                       0,                       0,                       0,                 0,                 0,                 0,  a_y - r_{(5,6)y}, -a_x + r_{(5,6)x},                 0],\n[                                                                                                                 0,                 0, -a_z + r_{(2,6)z},  a_y - r_{(2,6)y},                       0,        a_z - r_{(3,4)z},       -a_y + r_{(3,4)y},                 0,  a_z - r_{(4,6)z}, -a_y + r_{(4,6)y},                 0,                 0,                 0],\n[                                                                                                                 0,  a_z - r_{(2,6)z},                 0, -a_x + r_{(2,6)x},       -a_z + r_{(3,4)z},                       0,        a_x - r_{(3,4)x}, -a_z + r_{(4,6)z},                 0,  a_x - r_{(4,6)x},                 0,                 0,                 0],\n[                                                                                                                 0, -a_y + r_{(2,6)y},  a_x - r_{(2,6)x},                 0,        a_y - r_{(3,4)y},       -a_x + r_{(3,4)x},                       0,  a_y - r_{(4,6)y}, -a_x + r_{(4,6)x},                 0,                 0,                 0,                 0],\n[                                                                              sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}),                -1,                 0,                 0,                       0,                       0,                       0,                 0,                 0,                 0,                 1,                 0,                 0],\n[                                                                              sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}),                 0,                -1,                 0,                       0,                       0,                       0,                 0,                 0,                 0,                 0,                 1,                 0],\n[                                                                                                cos(\\beta_{(1,5)}),                 0,                 0,                -1,                       0,                       0,                       0,                 0,                 0,                 0,                 0,                 0,                 1],\n[                                                                                                                 0,                -1,                 0,                 0,                       1,                       0,                       0,                 1,                 0,                 0,                 0,                 0,                 0],\n[                                                                                                                 0,                 0,                -1,                 0,                       0,                       1,                       0,                 0,                 1,                 0,                 0,                 0,                 0],\n[                                                                                                                 0,                 0,                 0,                -1,                       0,                       0,                       1,                 0,                 0,                 1,                 0,                 0,                 0],\n[                                                                                                                 0,                 0,                 0,                 0, r_{(3,4)x} - r_{(4,6)x}, r_{(3,4)y} - r_{(4,6)y}, r_{(3,4)z} - r_{(4,6)z},                 0,                 0,                 0,                 0,                 0,                 0]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[-0.002504146920066251, -0.07609560862669454], [0.012075742999220117, 0.07032746303259974], [0.03148123588243843, 0.0597912455704322], [0.5953123975144231, -0.007041488792730323], [0.3297665534987283, -0.1401063221729949], [0.8201972001843377, 0.15583378016365668]])).all()


def test_spatial_6UPS():
    A = numpy.array(
        [[9, 5, 0, 5, 0, 5, 0, 5, 0, 5, 0, 5, 0, 0],
         [0, 9, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 1, 9, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 4],
         [0, 0, 0, 9, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 1, 9, 0, 0, 0, 0, 0, 0, 0, 0, 4],
         [0, 0, 0, 0, 0, 9, 2, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 1, 9, 0, 0, 0, 0, 0, 0, 4],
         [0, 0, 0, 0, 0, 0, 0, 9, 2, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 1, 9, 0, 0, 0, 0, 4],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 2, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 9, 0, 0, 4],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 2, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 9, 4],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9]]
    )
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}), 0, 0, 0, 0, 0],\n[sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}), 0, 0, 0, 0, 0],\n[                  cos(\\beta_{(2,3)}), 0, 0, 0, 0, 0],\n[                                   0, 0, 0, 0, 0, 0],\n[                                   0, 0, 0, 0, 0, 0],\n[                                   0, 0, 0, 0, 0, 0]]), Matrix([\n[                  -(a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                            -(a_y - r_{(1,2)y})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) - (a_z - r_{(1,2)z})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                  0,  a_z - r_{(3,14)z}, -a_y + r_{(3,14)y}, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                   (a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                             (a_x - r_{(1,2)x})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) + (a_z - r_{(1,2)z})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -a_z + r_{(3,14)z},                  0,  a_x - r_{(3,14)x}, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[-(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), (a_x - r_{(1,2)x})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})) - (a_y - r_{(1,2)y})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,  a_y - r_{(3,14)y}, -a_x + r_{(3,14)x},                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                                                              sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                                                                                              -cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                  1,                  0,                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                                                              sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                                                                                              -sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                  0,                  1,                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                                                                                cos(\\beta_{(1,2)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                  0,                  0,                  1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]), Matrix([\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}), sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                                    0,                                    0,                                        0,                                        0],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}), sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                                    0,                                    0,                                        0,                                        0],\n[                  -cos(\\beta_{(2,3)}),                   cos(\\beta_{(4,5)}),                                    0,                                    0,                                        0,                                        0],\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                                    0, sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                                    0,                                        0,                                        0],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                                    0, sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                                    0,                                        0,                                        0],\n[                  -cos(\\beta_{(2,3)}),                                    0,                   cos(\\beta_{(6,7)}),                                    0,                                        0,                                        0],\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                                    0,                                    0, sin(\\beta_{(8,9)})*cos(\\phi_{(8,9)}),                                        0,                                        0],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                                    0,                                    0, sin(\\beta_{(8,9)})*sin(\\phi_{(8,9)}),                                        0,                                        0],\n[                  -cos(\\beta_{(2,3)}),                                    0,                                    0,                   cos(\\beta_{(8,9)}),                                        0,                                        0],\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                                    0,                                    0,                                    0, sin(\\beta_{(10,11)})*cos(\\phi_{(10,11)}),                                        0],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                                    0,                                    0,                                    0, sin(\\beta_{(10,11)})*sin(\\phi_{(10,11)}),                                        0],\n[                  -cos(\\beta_{(2,3)}),                                    0,                                    0,                                    0,                     cos(\\beta_{(10,11)}),                                        0],\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                                    0,                                    0,                                    0,                                        0, sin(\\beta_{(12,13)})*cos(\\phi_{(12,13)})],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                                    0,                                    0,                                    0,                                        0, sin(\\beta_{(12,13)})*sin(\\phi_{(12,13)})],\n[                  -cos(\\beta_{(2,3)}),                                    0,                                    0,                                    0,                                        0,                     cos(\\beta_{(12,13)})],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                    0,                                        0,                                        0]]), Matrix([\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                              (a_y - r_{(1,2)y})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) + (a_z - r_{(1,2)z})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                   -(a_y - r_{(1,4)y})*cos(\\beta_{(1,4)}) + (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                                                                            -(a_y - r_{(1,4)y})/sqrt(1 + 1/(cos(\\delta_{(1,4)} - \\phi_{(1,4)})**2*tan(\\beta_{(1,4)})**2)) - (a_z - r_{(1,4)z})*sin(\\delta_{(1,4)})/(sqrt(1 + 1/(cos(\\delta_{(1,4)} - \\phi_{(1,4)})**2*tan(\\beta_{(1,4)})**2))*cos(\\delta_{(1,4)} - \\phi_{(1,4)})*tan(\\beta_{(1,4)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0, -a_z + r_{(3,14)z},  a_y - r_{(3,14)y},                  0,  a_z - r_{(5,14)z}, -a_y + r_{(5,14)y},                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                             -(a_x - r_{(1,2)x})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) - (a_z - r_{(1,2)z})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                    (a_x - r_{(1,4)x})*cos(\\beta_{(1,4)}) - (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                                                             (a_x - r_{(1,4)x})/sqrt(1 + 1/(cos(\\delta_{(1,4)} - \\phi_{(1,4)})**2*tan(\\beta_{(1,4)})**2)) + (a_z - r_{(1,4)z})*cos(\\delta_{(1,4)})/(sqrt(1 + 1/(cos(\\delta_{(1,4)} - \\phi_{(1,4)})**2*tan(\\beta_{(1,4)})**2))*cos(\\delta_{(1,4)} - \\phi_{(1,4)})*tan(\\beta_{(1,4)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,  a_z - r_{(3,14)z},                  0, -a_x + r_{(3,14)x}, -a_z + r_{(5,14)z},                  0,  a_x - r_{(5,14)x},                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -(a_x - r_{(1,2)x})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})) + (a_y - r_{(1,2)y})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})), -(a_x - r_{(1,4)x})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}) + (a_y - r_{(1,4)y})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}), (a_x - r_{(1,4)x})*sin(\\delta_{(1,4)})/(sqrt(1 + 1/(cos(\\delta_{(1,4)} - \\phi_{(1,4)})**2*tan(\\beta_{(1,4)})**2))*cos(\\delta_{(1,4)} - \\phi_{(1,4)})*tan(\\beta_{(1,4)})) - (a_y - r_{(1,4)y})*cos(\\delta_{(1,4)})/(sqrt(1 + 1/(cos(\\delta_{(1,4)} - \\phi_{(1,4)})**2*tan(\\beta_{(1,4)})**2))*cos(\\delta_{(1,4)} - \\phi_{(1,4)})*tan(\\beta_{(1,4)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0, -a_y + r_{(3,14)y},  a_x - r_{(3,14)x},                  0,  a_y - r_{(5,14)y}, -a_x + r_{(5,14)x},                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                              (a_y - r_{(1,2)y})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) + (a_z - r_{(1,2)z})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                   -(a_y - r_{(1,6)y})*cos(\\beta_{(1,6)}) + (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                                                                            -(a_y - r_{(1,6)y})/sqrt(1 + 1/(cos(\\delta_{(1,6)} - \\phi_{(1,6)})**2*tan(\\beta_{(1,6)})**2)) - (a_z - r_{(1,6)z})*sin(\\delta_{(1,6)})/(sqrt(1 + 1/(cos(\\delta_{(1,6)} - \\phi_{(1,6)})**2*tan(\\beta_{(1,6)})**2))*cos(\\delta_{(1,6)} - \\phi_{(1,6)})*tan(\\beta_{(1,6)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0, -a_z + r_{(3,14)z},  a_y - r_{(3,14)y},                  0,                  0,                  0,                  0,  a_z - r_{(7,14)z}, -a_y + r_{(7,14)y},                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                             -(a_x - r_{(1,2)x})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) - (a_z - r_{(1,2)z})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                    (a_x - r_{(1,6)x})*cos(\\beta_{(1,6)}) - (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                                                                             (a_x - r_{(1,6)x})/sqrt(1 + 1/(cos(\\delta_{(1,6)} - \\phi_{(1,6)})**2*tan(\\beta_{(1,6)})**2)) + (a_z - r_{(1,6)z})*cos(\\delta_{(1,6)})/(sqrt(1 + 1/(cos(\\delta_{(1,6)} - \\phi_{(1,6)})**2*tan(\\beta_{(1,6)})**2))*cos(\\delta_{(1,6)} - \\phi_{(1,6)})*tan(\\beta_{(1,6)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,  a_z - r_{(3,14)z},                  0, -a_x + r_{(3,14)x},                  0,                  0,                  0, -a_z + r_{(7,14)z},                  0,  a_x - r_{(7,14)x},                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -(a_x - r_{(1,2)x})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})) + (a_y - r_{(1,2)y})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0, -(a_x - r_{(1,6)x})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}) + (a_y - r_{(1,6)y})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}), (a_x - r_{(1,6)x})*sin(\\delta_{(1,6)})/(sqrt(1 + 1/(cos(\\delta_{(1,6)} - \\phi_{(1,6)})**2*tan(\\beta_{(1,6)})**2))*cos(\\delta_{(1,6)} - \\phi_{(1,6)})*tan(\\beta_{(1,6)})) - (a_y - r_{(1,6)y})*cos(\\delta_{(1,6)})/(sqrt(1 + 1/(cos(\\delta_{(1,6)} - \\phi_{(1,6)})**2*tan(\\beta_{(1,6)})**2))*cos(\\delta_{(1,6)} - \\phi_{(1,6)})*tan(\\beta_{(1,6)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0, -a_y + r_{(3,14)y},  a_x - r_{(3,14)x},                  0,                  0,                  0,                  0,  a_y - r_{(7,14)y}, -a_x + r_{(7,14)x},                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                              (a_y - r_{(1,2)y})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) + (a_z - r_{(1,2)z})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                   -(a_y - r_{(1,8)y})*cos(\\beta_{(1,8)}) + (a_z - r_{(1,8)z})*sin(\\beta_{(1,8)})*sin(\\phi_{(1,8)}),                                                                            -(a_y - r_{(1,8)y})/sqrt(1 + 1/(cos(\\delta_{(1,8)} - \\phi_{(1,8)})**2*tan(\\beta_{(1,8)})**2)) - (a_z - r_{(1,8)z})*sin(\\delta_{(1,8)})/(sqrt(1 + 1/(cos(\\delta_{(1,8)} - \\phi_{(1,8)})**2*tan(\\beta_{(1,8)})**2))*cos(\\delta_{(1,8)} - \\phi_{(1,8)})*tan(\\beta_{(1,8)})),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0, -a_z + r_{(3,14)z},  a_y - r_{(3,14)y},                  0,                  0,                  0,                  0,                  0,                  0,                  0,  a_z - r_{(9,14)z}, -a_y + r_{(9,14)y},                   0,                   0,                   0,                   0,                   0,                   0],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                             -(a_x - r_{(1,2)x})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) - (a_z - r_{(1,2)z})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                    (a_x - r_{(1,8)x})*cos(\\beta_{(1,8)}) - (a_z - r_{(1,8)z})*sin(\\beta_{(1,8)})*cos(\\phi_{(1,8)}),                                                                             (a_x - r_{(1,8)x})/sqrt(1 + 1/(cos(\\delta_{(1,8)} - \\phi_{(1,8)})**2*tan(\\beta_{(1,8)})**2)) + (a_z - r_{(1,8)z})*cos(\\delta_{(1,8)})/(sqrt(1 + 1/(cos(\\delta_{(1,8)} - \\phi_{(1,8)})**2*tan(\\beta_{(1,8)})**2))*cos(\\delta_{(1,8)} - \\phi_{(1,8)})*tan(\\beta_{(1,8)})),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,  a_z - r_{(3,14)z},                  0, -a_x + r_{(3,14)x},                  0,                  0,                  0,                  0,                  0,                  0, -a_z + r_{(9,14)z},                  0,  a_x - r_{(9,14)x},                   0,                   0,                   0,                   0,                   0,                   0],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -(a_x - r_{(1,2)x})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})) + (a_y - r_{(1,2)y})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0, -(a_x - r_{(1,8)x})*sin(\\beta_{(1,8)})*sin(\\phi_{(1,8)}) + (a_y - r_{(1,8)y})*sin(\\beta_{(1,8)})*cos(\\phi_{(1,8)}), (a_x - r_{(1,8)x})*sin(\\delta_{(1,8)})/(sqrt(1 + 1/(cos(\\delta_{(1,8)} - \\phi_{(1,8)})**2*tan(\\beta_{(1,8)})**2))*cos(\\delta_{(1,8)} - \\phi_{(1,8)})*tan(\\beta_{(1,8)})) - (a_y - r_{(1,8)y})*cos(\\delta_{(1,8)})/(sqrt(1 + 1/(cos(\\delta_{(1,8)} - \\phi_{(1,8)})**2*tan(\\beta_{(1,8)})**2))*cos(\\delta_{(1,8)} - \\phi_{(1,8)})*tan(\\beta_{(1,8)})),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0, -a_y + r_{(3,14)y},  a_x - r_{(3,14)x},                  0,                  0,                  0,                  0,                  0,                  0,                  0,  a_y - r_{(9,14)y}, -a_x + r_{(9,14)x},                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                              (a_y - r_{(1,2)y})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) + (a_z - r_{(1,2)z})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                    -(a_y - r_{(1,10)y})*cos(\\beta_{(1,10)}) + (a_z - r_{(1,10)z})*sin(\\beta_{(1,10)})*sin(\\phi_{(1,10)}),                                                                                -(a_y - r_{(1,10)y})/sqrt(1 + 1/(cos(\\delta_{(1,10)} - \\phi_{(1,10)})**2*tan(\\beta_{(1,10)})**2)) - (a_z - r_{(1,10)z})*sin(\\delta_{(1,10)})/(sqrt(1 + 1/(cos(\\delta_{(1,10)} - \\phi_{(1,10)})**2*tan(\\beta_{(1,10)})**2))*cos(\\delta_{(1,10)} - \\phi_{(1,10)})*tan(\\beta_{(1,10)})),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0, -a_z + r_{(3,14)z},  a_y - r_{(3,14)y},                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,  a_z - r_{(11,14)z}, -a_y + r_{(11,14)y},                   0,                   0,                   0],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                             -(a_x - r_{(1,2)x})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) - (a_z - r_{(1,2)z})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                     (a_x - r_{(1,10)x})*cos(\\beta_{(1,10)}) - (a_z - r_{(1,10)z})*sin(\\beta_{(1,10)})*cos(\\phi_{(1,10)}),                                                                                 (a_x - r_{(1,10)x})/sqrt(1 + 1/(cos(\\delta_{(1,10)} - \\phi_{(1,10)})**2*tan(\\beta_{(1,10)})**2)) + (a_z - r_{(1,10)z})*cos(\\delta_{(1,10)})/(sqrt(1 + 1/(cos(\\delta_{(1,10)} - \\phi_{(1,10)})**2*tan(\\beta_{(1,10)})**2))*cos(\\delta_{(1,10)} - \\phi_{(1,10)})*tan(\\beta_{(1,10)})),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,  a_z - r_{(3,14)z},                  0, -a_x + r_{(3,14)x},                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0, -a_z + r_{(11,14)z},                   0,  a_x - r_{(11,14)x},                   0,                   0,                   0],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -(a_x - r_{(1,2)x})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})) + (a_y - r_{(1,2)y})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0, -(a_x - r_{(1,10)x})*sin(\\beta_{(1,10)})*sin(\\phi_{(1,10)}) + (a_y - r_{(1,10)y})*sin(\\beta_{(1,10)})*cos(\\phi_{(1,10)}), (a_x - r_{(1,10)x})*sin(\\delta_{(1,10)})/(sqrt(1 + 1/(cos(\\delta_{(1,10)} - \\phi_{(1,10)})**2*tan(\\beta_{(1,10)})**2))*cos(\\delta_{(1,10)} - \\phi_{(1,10)})*tan(\\beta_{(1,10)})) - (a_y - r_{(1,10)y})*cos(\\delta_{(1,10)})/(sqrt(1 + 1/(cos(\\delta_{(1,10)} - \\phi_{(1,10)})**2*tan(\\beta_{(1,10)})**2))*cos(\\delta_{(1,10)} - \\phi_{(1,10)})*tan(\\beta_{(1,10)})),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0, -a_y + r_{(3,14)y},  a_x - r_{(3,14)x},                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,  a_y - r_{(11,14)y}, -a_x + r_{(11,14)x},                   0,                   0,                   0,                   0],\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                              (a_y - r_{(1,2)y})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) + (a_z - r_{(1,2)z})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                    -(a_y - r_{(1,12)y})*cos(\\beta_{(1,12)}) + (a_z - r_{(1,12)z})*sin(\\beta_{(1,12)})*sin(\\phi_{(1,12)}),                                                                                -(a_y - r_{(1,12)y})/sqrt(1 + 1/(cos(\\delta_{(1,12)} - \\phi_{(1,12)})**2*tan(\\beta_{(1,12)})**2)) - (a_z - r_{(1,12)z})*sin(\\delta_{(1,12)})/(sqrt(1 + 1/(cos(\\delta_{(1,12)} - \\phi_{(1,12)})**2*tan(\\beta_{(1,12)})**2))*cos(\\delta_{(1,12)} - \\phi_{(1,12)})*tan(\\beta_{(1,12)})),                  0, -a_z + r_{(3,14)z},  a_y - r_{(3,14)y},                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,  a_z - r_{(13,14)z}, -a_y + r_{(13,14)y}],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                             -(a_x - r_{(1,2)x})/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)) - (a_z - r_{(1,2)z})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                     (a_x - r_{(1,12)x})*cos(\\beta_{(1,12)}) - (a_z - r_{(1,12)z})*sin(\\beta_{(1,12)})*cos(\\phi_{(1,12)}),                                                                                 (a_x - r_{(1,12)x})/sqrt(1 + 1/(cos(\\delta_{(1,12)} - \\phi_{(1,12)})**2*tan(\\beta_{(1,12)})**2)) + (a_z - r_{(1,12)z})*cos(\\delta_{(1,12)})/(sqrt(1 + 1/(cos(\\delta_{(1,12)} - \\phi_{(1,12)})**2*tan(\\beta_{(1,12)})**2))*cos(\\delta_{(1,12)} - \\phi_{(1,12)})*tan(\\beta_{(1,12)})),  a_z - r_{(3,14)z},                  0, -a_x + r_{(3,14)x},                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0, -a_z + r_{(13,14)z},                   0,  a_x - r_{(13,14)x}],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -(a_x - r_{(1,2)x})*sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})) + (a_y - r_{(1,2)y})*cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0, -(a_x - r_{(1,12)x})*sin(\\beta_{(1,12)})*sin(\\phi_{(1,12)}) + (a_y - r_{(1,12)y})*sin(\\beta_{(1,12)})*cos(\\phi_{(1,12)}), (a_x - r_{(1,12)x})*sin(\\delta_{(1,12)})/(sqrt(1 + 1/(cos(\\delta_{(1,12)} - \\phi_{(1,12)})**2*tan(\\beta_{(1,12)})**2))*cos(\\delta_{(1,12)} - \\phi_{(1,12)})*tan(\\beta_{(1,12)})) - (a_y - r_{(1,12)y})*cos(\\delta_{(1,12)})/(sqrt(1 + 1/(cos(\\delta_{(1,12)} - \\phi_{(1,12)})**2*tan(\\beta_{(1,12)})**2))*cos(\\delta_{(1,12)} - \\phi_{(1,12)})*tan(\\beta_{(1,12)})), -a_y + r_{(3,14)y},  a_x - r_{(3,14)x},                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,  a_y - r_{(13,14)y}, -a_x + r_{(13,14)x},                   0],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                                                                                                cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                               sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                                                                                                                                                                              -cos(\\delta_{(1,4)})/(sqrt(1 + 1/(cos(\\delta_{(1,4)} - \\phi_{(1,4)})**2*tan(\\beta_{(1,4)})**2))*cos(\\delta_{(1,4)} - \\phi_{(1,4)})*tan(\\beta_{(1,4)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                 -1,                  0,                  0,                  1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                                                                                                sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                               sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                                                                                                                                                                                              -sin(\\delta_{(1,4)})/(sqrt(1 + 1/(cos(\\delta_{(1,4)} - \\phi_{(1,4)})**2*tan(\\beta_{(1,4)})**2))*cos(\\delta_{(1,4)} - \\phi_{(1,4)})*tan(\\beta_{(1,4)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0,                 -1,                  0,                  0,                  1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                                                                                                                                                                                         -1/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)),                                                                                                 cos(\\beta_{(1,4)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(1,4)} - \\phi_{(1,4)})**2*tan(\\beta_{(1,4)})**2)),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0,                  0,                 -1,                  0,                  0,                  1,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                                                                                                cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                               sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                                                                                                                                                                                              -cos(\\delta_{(1,6)})/(sqrt(1 + 1/(cos(\\delta_{(1,6)} - \\phi_{(1,6)})**2*tan(\\beta_{(1,6)})**2))*cos(\\delta_{(1,6)} - \\phi_{(1,6)})*tan(\\beta_{(1,6)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                 -1,                  0,                  0,                  0,                  0,                  0,                  1,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                                                                                                sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                               sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                                                                                                                                                                                              -sin(\\delta_{(1,6)})/(sqrt(1 + 1/(cos(\\delta_{(1,6)} - \\phi_{(1,6)})**2*tan(\\beta_{(1,6)})**2))*cos(\\delta_{(1,6)} - \\phi_{(1,6)})*tan(\\beta_{(1,6)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0,                 -1,                  0,                  0,                  0,                  0,                  0,                  1,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                                                                                                                                                                                         -1/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                 cos(\\beta_{(1,6)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(1,6)} - \\phi_{(1,6)})**2*tan(\\beta_{(1,6)})**2)),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0,                  0,                 -1,                  0,                  0,                  0,                  0,                  0,                  1,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                                                                                                cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                               sin(\\beta_{(1,8)})*cos(\\phi_{(1,8)}),                                                                                                                                                                                              -cos(\\delta_{(1,8)})/(sqrt(1 + 1/(cos(\\delta_{(1,8)} - \\phi_{(1,8)})**2*tan(\\beta_{(1,8)})**2))*cos(\\delta_{(1,8)} - \\phi_{(1,8)})*tan(\\beta_{(1,8)})),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                 -1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  1,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                                                                                                sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                               sin(\\beta_{(1,8)})*sin(\\phi_{(1,8)}),                                                                                                                                                                                              -sin(\\delta_{(1,8)})/(sqrt(1 + 1/(cos(\\delta_{(1,8)} - \\phi_{(1,8)})**2*tan(\\beta_{(1,8)})**2))*cos(\\delta_{(1,8)} - \\phi_{(1,8)})*tan(\\beta_{(1,8)})),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0,                 -1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  1,                  0,                   0,                   0,                   0,                   0,                   0,                   0],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                                                                                                                                                                                         -1/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                 cos(\\beta_{(1,8)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(1,8)} - \\phi_{(1,8)})**2*tan(\\beta_{(1,8)})**2)),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0,                  0,                 -1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  1,                   0,                   0,                   0,                   0,                   0,                   0],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                                                                                                cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                   sin(\\beta_{(1,10)})*cos(\\phi_{(1,10)}),                                                                                                                                                                                                       -cos(\\delta_{(1,10)})/(sqrt(1 + 1/(cos(\\delta_{(1,10)} - \\phi_{(1,10)})**2*tan(\\beta_{(1,10)})**2))*cos(\\delta_{(1,10)} - \\phi_{(1,10)})*tan(\\beta_{(1,10)})),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                 -1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   1,                   0,                   0,                   0,                   0,                   0],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                                                                                                sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                   sin(\\beta_{(1,10)})*sin(\\phi_{(1,10)}),                                                                                                                                                                                                       -sin(\\delta_{(1,10)})/(sqrt(1 + 1/(cos(\\delta_{(1,10)} - \\phi_{(1,10)})**2*tan(\\beta_{(1,10)})**2))*cos(\\delta_{(1,10)} - \\phi_{(1,10)})*tan(\\beta_{(1,10)})),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0,                 -1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   1,                   0,                   0,                   0,                   0],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                                                                                                                                                                                         -1/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                      cos(\\beta_{(1,10)}),                                                                                                                                                                                                                                                                                      1/sqrt(1 + 1/(cos(\\delta_{(1,10)} - \\phi_{(1,10)})**2*tan(\\beta_{(1,10)})**2)),                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                  0,                  0,                 -1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   1,                   0,                   0,                   0],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                                                                                                cos(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                   sin(\\beta_{(1,12)})*cos(\\phi_{(1,12)}),                                                                                                                                                                                                       -cos(\\delta_{(1,12)})/(sqrt(1 + 1/(cos(\\delta_{(1,12)} - \\phi_{(1,12)})**2*tan(\\beta_{(1,12)})**2))*cos(\\delta_{(1,12)} - \\phi_{(1,12)})*tan(\\beta_{(1,12)})),                 -1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   1,                   0,                   0],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                                                                                                sin(\\delta_{(1,2)})/(sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2))*cos(\\delta_{(1,2)} - \\phi_{(1,2)})*tan(\\beta_{(1,2)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                   sin(\\beta_{(1,12)})*sin(\\phi_{(1,12)}),                                                                                                                                                                                                       -sin(\\delta_{(1,12)})/(sqrt(1 + 1/(cos(\\delta_{(1,12)} - \\phi_{(1,12)})**2*tan(\\beta_{(1,12)})**2))*cos(\\delta_{(1,12)} - \\phi_{(1,12)})*tan(\\beta_{(1,12)})),                  0,                 -1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   1,                   0],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                                                                                                                                                                                         -1/sqrt(1 + 1/(cos(\\delta_{(1,2)} - \\phi_{(1,2)})**2*tan(\\beta_{(1,2)})**2)),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                                        0,                                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                      cos(\\beta_{(1,12)}),                                                                                                                                                                                                                                                                                      1/sqrt(1 + 1/(cos(\\delta_{(1,12)} - \\phi_{(1,12)})**2*tan(\\beta_{(1,12)})**2)),                  0,                  0,                 -1,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0,                   0,                   0,                   1]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[5.774753346150521, 97.82256949187843, -45.15557075547076, -90.51504703507577, 18.290945467718046, -4.272492807480118], [-10.22560067511352, -172.23661975202774, 79.26381326096951, 160.53203885351846, -32.20751405728672, 7.539022591730421], [10.881404385894768, 182.44559604325102, -84.18497302617371, -168.97595383513382, 34.114181616062325, -7.970736434140461], [41.01015525604739, 689.0751623849666, -319.55311899460924, -638.925076516943, 128.2829379900573, -30.817187964278823], [-12.983025372580219, -216.5484094554666, 100.3532688778645, 203.0870539185442, -41.2624725633418, 8.80907575094105], [-3.6289391002158644, -61.94681211898643, 27.04269070964366, 60.55404664096891, -11.913361625741286, 1.5617171905338516]])).all()

    
def test_spatial_SPM():
    A = numpy.array(
        [[9, 1, 0, 1, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 9, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 1, 9, 0, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 9, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 1, 9, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 9, 2, 0, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 1, 9, 4, 0, 0, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 9, 4, 0, 4, 0, 4, 0, 5, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 9, 2, 0, 0, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 1, 9, 0, 0, 0, 0, 0, 4],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 2, 0, 0, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 9, 0, 0, 0, 4],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 2, 0, 0],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 9, 0, 4],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 2],
         [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9]]
    )
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}), 0, 0, 0, 0, 0],\n[sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}), 0, 0, 0, 0, 0],\n[                  cos(\\beta_{(2,3)}), 0, 0, 0, 0, 0],\n[                                   0, 0, 0, 0, 0, 0],\n[                                   0, 0, 0, 0, 0, 0],\n[                                   0, 0, 0, 0, 0, 0]]), Matrix([\n[                  -(a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), 0, 0,                 0,  a_z - r_{(3,8)z}, -a_y + r_{(3,8)y}, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                    -(a_y - r_{(8,15)y})*cos(\\beta_{(8,15)}) + (a_z - r_{(8,15)z})*sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}),                                                                                -(a_y - r_{(8,15)y})/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)) - (a_z - r_{(8,15)z})*sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})), 0, 0, 0, 0, 0, 0, 0, 0, 0, sin(\\beta_{(15,16)})*cos(\\phi_{(15,16)})],\n[                   (a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0, -a_z + r_{(3,8)z},                 0,  a_x - r_{(3,8)x}, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                     (a_x - r_{(8,15)x})*cos(\\beta_{(8,15)}) - (a_z - r_{(8,15)z})*sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}),                                                                                 (a_x - r_{(8,15)x})/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)) + (a_z - r_{(8,15)z})*cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})), 0, 0, 0, 0, 0, 0, 0, 0, 0, sin(\\beta_{(15,16)})*sin(\\phi_{(15,16)})],\n[-(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0,  a_y - r_{(3,8)y}, -a_x + r_{(3,8)x},                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, -(a_x - r_{(8,15)x})*sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}) + (a_y - r_{(8,15)y})*sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}), (a_x - r_{(8,15)x})*sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})) - (a_y - r_{(8,15)y})*cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})), 0, 0, 0, 0, 0, 0, 0, 0, 0,                     cos(\\beta_{(15,16)})],\n[                                                                              sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0,                 1,                 0,                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                                                                                   sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}),                                                                                                                                                                                                       -cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})), 0, 0, 0, 0, 0, 0, 0, 0, 0,                                        0],\n[                                                                              sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), 0, 0,                 0,                 1,                 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                                                                                   sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}),                                                                                                                                                                                                       -sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})), 0, 0, 0, 0, 0, 0, 0, 0, 0,                                        0],\n[                                                                                                cos(\\beta_{(1,2)}), 0, 0,                 0,                 0,                 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                                                                                                      cos(\\beta_{(8,15)}),                                                                                                                                                                                                                                                                                      1/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)), 0, 0, 0, 0, 0, 0, 0, 0, 0,                                        0]]), Matrix([\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}), sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                                    0,                                      0,                                        0,                                        0],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}), sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                                    0,                                      0,                                        0,                                        0],\n[                  -cos(\\beta_{(2,3)}),                   cos(\\beta_{(4,5)}),                                    0,                                      0,                                        0,                                        0],\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                                    0, sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                                      0,                                        0,                                        0],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                                    0, sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                                      0,                                        0,                                        0],\n[                  -cos(\\beta_{(2,3)}),                                    0,                   cos(\\beta_{(6,7)}),                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0, sin(\\beta_{(9,10)})*cos(\\phi_{(9,10)}),                                        0,                                        0],\n[                                    0,                                    0,                                    0, sin(\\beta_{(9,10)})*sin(\\phi_{(9,10)}),                                        0,                                        0],\n[                                    0,                                    0,                                    0,                    cos(\\beta_{(9,10)}),                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0, sin(\\beta_{(11,12)})*cos(\\phi_{(11,12)}),                                        0],\n[                                    0,                                    0,                                    0,                                      0, sin(\\beta_{(11,12)})*sin(\\phi_{(11,12)}),                                        0],\n[                                    0,                                    0,                                    0,                                      0,                     cos(\\beta_{(11,12)}),                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0, sin(\\beta_{(13,14)})*cos(\\phi_{(13,14)})],\n[                                    0,                                    0,                                    0,                                      0,                                        0, sin(\\beta_{(13,14)})*sin(\\phi_{(13,14)})],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                     cos(\\beta_{(13,14)})],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0],\n[                                    0,                                    0,                                    0,                                      0,                                        0,                                        0]]), Matrix([\n[                                                                                           (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                   -(a_y - r_{(1,4)y})*cos(\\beta_{(1,4)}) + (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                                                                                                                  0,                           0,           -a_z + r_{(3,8)z},            a_y - r_{(3,8)y},                 0,  a_z - r_{(5,8)z}, -a_y + r_{(5,8)y},                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                          -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                    (a_x - r_{(1,4)x})*cos(\\beta_{(1,4)}) - (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                                                                                                  0,            a_z - r_{(3,8)z},                           0,           -a_x + r_{(3,8)x}, -a_z + r_{(5,8)z},                 0,  a_x - r_{(5,8)x},                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                         (a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -(a_x - r_{(1,4)x})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}) + (a_y - r_{(1,4)y})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                                                                                                  0,           -a_y + r_{(3,8)y},            a_x - r_{(3,8)x},                           0,  a_y - r_{(5,8)y}, -a_x + r_{(5,8)x},                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                           (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                  0,                   -(a_y - r_{(1,6)y})*cos(\\beta_{(1,6)}) + (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                           0,           -a_z + r_{(3,8)z},            a_y - r_{(3,8)y},                 0,                 0,                 0,                 0,  a_z - r_{(7,8)z}, -a_y + r_{(7,8)y},                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                          -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                  0,                    (a_x - r_{(1,6)x})*cos(\\beta_{(1,6)}) - (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),            a_z - r_{(3,8)z},                           0,           -a_x + r_{(3,8)x},                 0,                 0,                 0, -a_z + r_{(7,8)z},                 0,  a_x - r_{(7,8)x},                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                         (a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                  0, -(a_x - r_{(1,6)x})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}) + (a_y - r_{(1,6)y})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),           -a_y + r_{(3,8)y},            a_x - r_{(3,8)x},                           0,                 0,                 0,                 0,  a_y - r_{(7,8)y}, -a_x + r_{(7,8)x},                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,           a_z - r_{(8,9)z},          -a_y + r_{(8,9)y},                           0,                           0,                           0,                           0,                           0,                           0,                    (a_y - r_{(8,15)y})*cos(\\beta_{(8,15)}) - (a_z - r_{(8,15)z})*sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}),                                                                                  (a_y - r_{(8,15)y})/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)) + (a_z - r_{(8,15)z})*sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,  a_z - r_{(10,16)z}, -a_y + r_{(10,16)y},                   0,                   0,                   0,                   0,                   0,                   0, -sin(\\beta_{(15,16)})*cos(\\phi_{(15,16)})],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,          -a_z + r_{(8,9)z},                          0,           a_x - r_{(8,9)x},                           0,                           0,                           0,                           0,                           0,                           0,                   -(a_x - r_{(8,15)x})*cos(\\beta_{(8,15)}) + (a_z - r_{(8,15)z})*sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}),                                                                                 -(a_x - r_{(8,15)x})/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)) - (a_z - r_{(8,15)z})*cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})), -a_z + r_{(10,16)z},                   0,  a_x - r_{(10,16)x},                   0,                   0,                   0,                   0,                   0,                   0, -sin(\\beta_{(15,16)})*sin(\\phi_{(15,16)})],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,           a_y - r_{(8,9)y},          -a_x + r_{(8,9)x},                          0,                           0,                           0,                           0,                           0,                           0,                           0, (a_x - r_{(8,15)x})*sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}) - (a_y - r_{(8,15)y})*sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}), -(a_x - r_{(8,15)x})*sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})) + (a_y - r_{(8,15)y})*cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),  a_y - r_{(10,16)y}, -a_x + r_{(10,16)x},                   0,                   0,                   0,                   0,                   0,                   0,                   0,                     -cos(\\beta_{(15,16)})],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,           a_z - r_{(8,11)z},          -a_y + r_{(8,11)y},                           0,                           0,                           0,                    (a_y - r_{(8,15)y})*cos(\\beta_{(8,15)}) - (a_z - r_{(8,15)z})*sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}),                                                                                  (a_y - r_{(8,15)y})/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)) + (a_z - r_{(8,15)z})*sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   0,                   0,                   0,  a_z - r_{(12,16)z}, -a_y + r_{(12,16)y},                   0,                   0,                   0, -sin(\\beta_{(15,16)})*cos(\\phi_{(15,16)})],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,          -a_z + r_{(8,11)z},                           0,           a_x - r_{(8,11)x},                           0,                           0,                           0,                   -(a_x - r_{(8,15)x})*cos(\\beta_{(8,15)}) + (a_z - r_{(8,15)z})*sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}),                                                                                 -(a_x - r_{(8,15)x})/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)) - (a_z - r_{(8,15)z})*cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   0,                   0, -a_z + r_{(12,16)z},                   0,  a_x - r_{(12,16)x},                   0,                   0,                   0, -sin(\\beta_{(15,16)})*sin(\\phi_{(15,16)})],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,           a_y - r_{(8,11)y},          -a_x + r_{(8,11)x},                           0,                           0,                           0,                           0, (a_x - r_{(8,15)x})*sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}) - (a_y - r_{(8,15)y})*sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}), -(a_x - r_{(8,15)x})*sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})) + (a_y - r_{(8,15)y})*cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   0,                   0,  a_y - r_{(12,16)y}, -a_x + r_{(12,16)x},                   0,                   0,                   0,                   0,                     -cos(\\beta_{(15,16)})],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,           a_z - r_{(8,13)z},          -a_y + r_{(8,13)y},                    (a_y - r_{(8,15)y})*cos(\\beta_{(8,15)}) - (a_z - r_{(8,15)z})*sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}),                                                                                  (a_y - r_{(8,15)y})/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)) + (a_z - r_{(8,15)z})*sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   0,                   0,                   0,                   0,                   0,                   0,  a_z - r_{(14,16)z}, -a_y + r_{(14,16)y}, -sin(\\beta_{(15,16)})*cos(\\phi_{(15,16)})],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,          -a_z + r_{(8,13)z},                           0,           a_x - r_{(8,13)x},                   -(a_x - r_{(8,15)x})*cos(\\beta_{(8,15)}) + (a_z - r_{(8,15)z})*sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}),                                                                                 -(a_x - r_{(8,15)x})/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)) - (a_z - r_{(8,15)z})*cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   0,                   0,                   0,                   0,                   0, -a_z + r_{(14,16)z},                   0,  a_x - r_{(14,16)x}, -sin(\\beta_{(15,16)})*sin(\\phi_{(15,16)})],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,           a_y - r_{(8,13)y},          -a_x + r_{(8,13)x},                           0, (a_x - r_{(8,15)x})*sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}) - (a_y - r_{(8,15)y})*sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}), -(a_x - r_{(8,15)x})*sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})) + (a_y - r_{(8,15)y})*cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   0,                   0,                   0,                   0,                   0,  a_y - r_{(14,16)y}, -a_x + r_{(14,16)x},                   0,                     -cos(\\beta_{(15,16)})],\n[                                                                                                                                                     -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                               sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                                                                                                  0,                          -1,                           0,                           0,                 1,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                     -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                               sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                                                                                                                  0,                           0,                          -1,                           0,                 0,                 1,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                                       -cos(\\beta_{(1,2)}),                                                                                                 cos(\\beta_{(1,4)}),                                                                                                                  0,                           0,                           0,                          -1,                 0,                 0,                 1,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                     -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                  0,                                                                               sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                          -1,                           0,                           0,                 0,                 0,                 0,                 1,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                     -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                  0,                                                                               sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                           0,                          -1,                           0,                 0,                 0,                 0,                 0,                 1,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                                       -cos(\\beta_{(1,2)}),                                                                                                                  0,                                                                                                 cos(\\beta_{(1,6)}),                           0,                           0,                          -1,                 0,                 0,                 0,                 0,                 0,                 1,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          1,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                 -sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}),                                                                                                                                                                                                         cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   1,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          1,                          0,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                 -sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}),                                                                                                                                                                                                         sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   1,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          1,                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                    -cos(\\beta_{(8,15)}),                                                                                                                                                                                                                                                                                      -1/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)),                   0,                   0,                   1,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           1,                           0,                           0,                           0,                           0,                           0,                                                                                 -sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}),                                                                                                                                                                                                         cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   0,                   0,                   1,                   0,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           1,                           0,                           0,                           0,                           0,                                                                                 -sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}),                                                                                                                                                                                                         sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   0,                   0,                   0,                   1,                   0,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           1,                           0,                           0,                           0,                                                                                                    -cos(\\beta_{(8,15)}),                                                                                                                                                                                                                                                                                      -1/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)),                   0,                   0,                   0,                   0,                   0,                   1,                   0,                   0,                   0,                                         0],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           1,                           0,                           0,                                                                                 -sin(\\beta_{(8,15)})*cos(\\phi_{(8,15)}),                                                                                                                                                                                                         cos(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   0,                   0,                   0,                   0,                   0,                   1,                   0,                   0,                                         0],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           1,                           0,                                                                                 -sin(\\beta_{(8,15)})*sin(\\phi_{(8,15)}),                                                                                                                                                                                                         sin(\\delta_{(8,15)})/(sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2))*cos(\\delta_{(8,15)} - \\phi_{(8,15)})*tan(\\beta_{(8,15)})),                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   1,                   0,                                         0],\n[                                                                                                                                                                                         0,                                                                                                                  0,                                                                                                                  0,                           0,                           0,                           0,                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0,                           0,                           0,                           1,                                                                                                    -cos(\\beta_{(8,15)}),                                                                                                                                                                                                                                                                                      -1/sqrt(1 + 1/(cos(\\delta_{(8,15)} - \\phi_{(8,15)})**2*tan(\\beta_{(8,15)})**2)),                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   1,                                         0],\n[   (-r_{(10,16)x} + r_{(8,9)x})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}) + (-r_{(10,16)y} + r_{(8,9)y})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (-r_{(10,16)z} + r_{(8,9)z})*cos(\\beta_{(1,2)}),                                                                                                                  0,                                                                                                                  0,  -r_{(10,16)x} + r_{(8,9)x},  -r_{(10,16)y} + r_{(8,9)y},  -r_{(10,16)z} + r_{(8,9)z},                 0,                 0,                 0,                 0,                 0,                 0, -r_{(10,16)x} + r_{(8,9)x}, -r_{(10,16)y} + r_{(8,9)y}, -r_{(10,16)z} + r_{(8,9)z},                           0,                           0,                           0,                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[(-r_{(12,16)x} + r_{(8,11)x})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}) + (-r_{(12,16)y} + r_{(8,11)y})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (-r_{(12,16)z} + r_{(8,11)z})*cos(\\beta_{(1,2)}),                                                                                                                  0,                                                                                                                  0, -r_{(12,16)x} + r_{(8,11)x}, -r_{(12,16)y} + r_{(8,11)y}, -r_{(12,16)z} + r_{(8,11)z},                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0, -r_{(12,16)x} + r_{(8,11)x}, -r_{(12,16)y} + r_{(8,11)y}, -r_{(12,16)z} + r_{(8,11)z},                           0,                           0,                           0,                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0],\n[(-r_{(14,16)x} + r_{(8,13)x})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}) + (-r_{(14,16)y} + r_{(8,13)y})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (-r_{(14,16)z} + r_{(8,13)z})*cos(\\beta_{(1,2)}),                                                                                                                  0,                                                                                                                  0, -r_{(14,16)x} + r_{(8,13)x}, -r_{(14,16)y} + r_{(8,13)y}, -r_{(14,16)z} + r_{(8,13)z},                 0,                 0,                 0,                 0,                 0,                 0,                          0,                          0,                          0,                           0,                           0,                           0, -r_{(14,16)x} + r_{(8,13)x}, -r_{(14,16)y} + r_{(8,13)y}, -r_{(14,16)z} + r_{(8,13)z},                                                                                                                       0,                                                                                                                                                                                                                                                                                                                                                                    0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                   0,                                         0]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[-0.2954830747091039, 0.16188535851545205, 0.42927881535478124, 0.34762181025618727, -0.0053435685541210086, -0.023263998529026723], [1.7318043041523092, -1.548789386276862, 0.363989053851306, 0.0902127322667511, 0.09040571912460622, -0.2874080260051061], [0.9947663535377226, 0.19300042104259987, -0.5416357488895045, -0.15631627932105394, -0.07562621562329297, 1.2894597941903725], [1.1315155386456357, 3.201554397206984, -3.8148117372912886, 0.15908553635662886, -0.5052752361109362, 1.330217292455864], [2.529836815247941, -1.367817468161494, -1.6516344425601917, 0.3961305553178086, -0.23100409249935164, 0.42594836715881806], [3.6255647815952723, -1.9816225528345492, -0.693545269171332, -0.9117459756886397, -0.03221584986852172, 0.6042157641718521]])).all()

    
def test_spatial_6CPU():
    A = numpy.array(
        [[9, 3, 0, 3, 0, 3, 0, 0],
         [0, 9, 2, 0, 0, 0, 0, 0],
         [0, 1, 9, 0, 0, 0, 0, 5],
         [0, 0, 0, 9, 2, 0, 0, 0],
         [0, 0, 0, 1, 9, 0, 0, 5],
         [0, 0, 0, 0, 0, 9, 2, 0],
         [0, 0, 0, 0, 0, 1, 9, 5],
         [0, 0, 0, 0, 0, 0, 0, 9]]
    )
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}), 0, 0],\n[sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}), 0, 0],\n[                  cos(\\beta_{(2,3)}), 0, 0],\n[                                   0, 0, 0],\n[                                   0, 0, 0],\n[                                   0, 0, 0]]), Matrix([\n[                  -(a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0, 0, 0,                   -(a_y - r_{(3,8)y})*cos(\\beta_{(3,8)}) + (a_z - r_{(3,8)z})*sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}),                                                                            -(a_y - r_{(3,8)y})/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)) - (a_z - r_{(3,8)z})*sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})), 0, 0, 0, 0],\n[                   (a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), 0, 0, 0, 0,                    (a_x - r_{(3,8)x})*cos(\\beta_{(3,8)}) - (a_z - r_{(3,8)z})*sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}),                                                                             (a_x - r_{(3,8)x})/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)) + (a_z - r_{(3,8)z})*cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})), 0, 0, 0, 0],\n[-(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                   cos(\\beta_{(1,2)}), 0, 0, 0, 0, -(a_x - r_{(3,8)x})*sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}) + (a_y - r_{(3,8)y})*sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}), (a_x - r_{(3,8)x})*sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})) - (a_y - r_{(3,8)y})*cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})), 0, 0, 0, 0],\n[                                                                              sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                    0, 0, 0, 0, 0,                                                                               sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}),                                                                                                                                                                                              -cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})), 0, 0, 0, 0],\n[                                                                              sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                    0, 0, 0, 0, 0,                                                                               sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}),                                                                                                                                                                                              -sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})), 0, 0, 0, 0],\n[                                                                                                cos(\\beta_{(1,2)}),                                    0, 0, 0, 0, 0,                                                                                                 cos(\\beta_{(3,8)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)), 0, 0, 0, 0]]), Matrix([\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}), sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                                    0],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}), sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                                    0],\n[                  -cos(\\beta_{(2,3)}),                   cos(\\beta_{(4,5)}),                                    0],\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                                    0, sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)})],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                                    0, sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)})],\n[                  -cos(\\beta_{(2,3)}),                                    0,                   cos(\\beta_{(6,7)})],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0]]), Matrix([\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                   -(a_y - r_{(1,4)y})*cos(\\beta_{(1,4)}) + (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}), sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                                                                                                  0,                                    0,                   (a_y - r_{(3,8)y})*cos(\\beta_{(3,8)}) - (a_z - r_{(3,8)z})*sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}),                                                                              (a_y - r_{(3,8)y})/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)) + (a_z - r_{(3,8)z})*sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                   -(a_y - r_{(5,8)y})*cos(\\beta_{(5,8)}) + (a_z - r_{(5,8)z})*sin(\\beta_{(5,8)})*sin(\\phi_{(5,8)}),                                                                            -(a_y - r_{(5,8)y})/sqrt(1 + 1/(cos(\\delta_{(5,8)} - \\phi_{(5,8)})**2*tan(\\beta_{(5,8)})**2)) - (a_z - r_{(5,8)z})*sin(\\delta_{(5,8)})/(sqrt(1 + 1/(cos(\\delta_{(5,8)} - \\phi_{(5,8)})**2*tan(\\beta_{(5,8)})**2))*cos(\\delta_{(5,8)} - \\phi_{(5,8)})*tan(\\beta_{(5,8)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                    (a_x - r_{(1,4)x})*cos(\\beta_{(1,4)}) - (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}), sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                                                                                                                  0,                                    0,                  -(a_x - r_{(3,8)x})*cos(\\beta_{(3,8)}) + (a_z - r_{(3,8)z})*sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}),                                                                             -(a_x - r_{(3,8)x})/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)) - (a_z - r_{(3,8)z})*cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                    (a_x - r_{(5,8)x})*cos(\\beta_{(5,8)}) - (a_z - r_{(5,8)z})*sin(\\beta_{(5,8)})*cos(\\phi_{(5,8)}),                                                                             (a_x - r_{(5,8)x})/sqrt(1 + 1/(cos(\\delta_{(5,8)} - \\phi_{(5,8)})**2*tan(\\beta_{(5,8)})**2)) + (a_z - r_{(5,8)z})*cos(\\delta_{(5,8)})/(sqrt(1 + 1/(cos(\\delta_{(5,8)} - \\phi_{(5,8)})**2*tan(\\beta_{(5,8)})**2))*cos(\\delta_{(5,8)} - \\phi_{(5,8)})*tan(\\beta_{(5,8)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                   -cos(\\beta_{(1,2)}), -(a_x - r_{(1,4)x})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}) + (a_y - r_{(1,4)y})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                   cos(\\beta_{(1,4)}),                                                                                                                  0,                                    0, (a_x - r_{(3,8)x})*sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}) - (a_y - r_{(3,8)y})*sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}), -(a_x - r_{(3,8)x})*sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})) + (a_y - r_{(3,8)y})*cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})), -(a_x - r_{(5,8)x})*sin(\\beta_{(5,8)})*sin(\\phi_{(5,8)}) + (a_y - r_{(5,8)y})*sin(\\beta_{(5,8)})*cos(\\phi_{(5,8)}), (a_x - r_{(5,8)x})*sin(\\delta_{(5,8)})/(sqrt(1 + 1/(cos(\\delta_{(5,8)} - \\phi_{(5,8)})**2*tan(\\beta_{(5,8)})**2))*cos(\\delta_{(5,8)} - \\phi_{(5,8)})*tan(\\beta_{(5,8)})) - (a_y - r_{(5,8)y})*cos(\\delta_{(5,8)})/(sqrt(1 + 1/(cos(\\delta_{(5,8)} - \\phi_{(5,8)})**2*tan(\\beta_{(5,8)})**2))*cos(\\delta_{(5,8)} - \\phi_{(5,8)})*tan(\\beta_{(5,8)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0],\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                  0,                                    0,                   -(a_y - r_{(1,6)y})*cos(\\beta_{(1,6)}) + (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}), sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                   (a_y - r_{(3,8)y})*cos(\\beta_{(3,8)}) - (a_z - r_{(3,8)z})*sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}),                                                                              (a_y - r_{(3,8)y})/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)) + (a_z - r_{(3,8)z})*sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                   -(a_y - r_{(7,8)y})*cos(\\beta_{(7,8)}) + (a_z - r_{(7,8)z})*sin(\\beta_{(7,8)})*sin(\\phi_{(7,8)}),                                                                            -(a_y - r_{(7,8)y})/sqrt(1 + 1/(cos(\\delta_{(7,8)} - \\phi_{(7,8)})**2*tan(\\beta_{(7,8)})**2)) - (a_z - r_{(7,8)z})*sin(\\delta_{(7,8)})/(sqrt(1 + 1/(cos(\\delta_{(7,8)} - \\phi_{(7,8)})**2*tan(\\beta_{(7,8)})**2))*cos(\\delta_{(7,8)} - \\phi_{(7,8)})*tan(\\beta_{(7,8)}))],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                  0,                                    0,                    (a_x - r_{(1,6)x})*cos(\\beta_{(1,6)}) - (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}), sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                  -(a_x - r_{(3,8)x})*cos(\\beta_{(3,8)}) + (a_z - r_{(3,8)z})*sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}),                                                                             -(a_x - r_{(3,8)x})/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)) - (a_z - r_{(3,8)z})*cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                    (a_x - r_{(7,8)x})*cos(\\beta_{(7,8)}) - (a_z - r_{(7,8)z})*sin(\\beta_{(7,8)})*cos(\\phi_{(7,8)}),                                                                             (a_x - r_{(7,8)x})/sqrt(1 + 1/(cos(\\delta_{(7,8)} - \\phi_{(7,8)})**2*tan(\\beta_{(7,8)})**2)) + (a_z - r_{(7,8)z})*cos(\\delta_{(7,8)})/(sqrt(1 + 1/(cos(\\delta_{(7,8)} - \\phi_{(7,8)})**2*tan(\\beta_{(7,8)})**2))*cos(\\delta_{(7,8)} - \\phi_{(7,8)})*tan(\\beta_{(7,8)}))],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                   -cos(\\beta_{(1,2)}),                                                                                                                  0,                                    0, -(a_x - r_{(1,6)x})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}) + (a_y - r_{(1,6)y})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                   cos(\\beta_{(1,6)}), (a_x - r_{(3,8)x})*sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}) - (a_y - r_{(3,8)y})*sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}), -(a_x - r_{(3,8)x})*sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})) + (a_y - r_{(3,8)y})*cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0, -(a_x - r_{(7,8)x})*sin(\\beta_{(7,8)})*sin(\\phi_{(7,8)}) + (a_y - r_{(7,8)y})*sin(\\beta_{(7,8)})*cos(\\phi_{(7,8)}), (a_x - r_{(7,8)x})*sin(\\delta_{(7,8)})/(sqrt(1 + 1/(cos(\\delta_{(7,8)} - \\phi_{(7,8)})**2*tan(\\beta_{(7,8)})**2))*cos(\\delta_{(7,8)} - \\phi_{(7,8)})*tan(\\beta_{(7,8)})) - (a_y - r_{(7,8)y})*cos(\\delta_{(7,8)})/(sqrt(1 + 1/(cos(\\delta_{(7,8)} - \\phi_{(7,8)})**2*tan(\\beta_{(7,8)})**2))*cos(\\delta_{(7,8)} - \\phi_{(7,8)})*tan(\\beta_{(7,8)}))],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                     0,                                                                               sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                    0,                                                                                                                  0,                                    0,                                                                             -sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}),                                                                                                                                                                                                cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                                                                               sin(\\beta_{(5,8)})*cos(\\phi_{(5,8)}),                                                                                                                                                                                              -cos(\\delta_{(5,8)})/(sqrt(1 + 1/(cos(\\delta_{(5,8)} - \\phi_{(5,8)})**2*tan(\\beta_{(5,8)})**2))*cos(\\delta_{(5,8)} - \\phi_{(5,8)})*tan(\\beta_{(5,8)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                     0,                                                                               sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                                    0,                                                                                                                  0,                                    0,                                                                             -sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}),                                                                                                                                                                                                sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                                                                               sin(\\beta_{(5,8)})*sin(\\phi_{(5,8)}),                                                                                                                                                                                              -sin(\\delta_{(5,8)})/(sqrt(1 + 1/(cos(\\delta_{(5,8)} - \\phi_{(5,8)})**2*tan(\\beta_{(5,8)})**2))*cos(\\delta_{(5,8)} - \\phi_{(5,8)})*tan(\\beta_{(5,8)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0],\n[                                                                                              -cos(\\beta_{(1,2)}),                                     0,                                                                                                 cos(\\beta_{(1,4)}),                                    0,                                                                                                                  0,                                    0,                                                                                               -cos(\\beta_{(3,8)}),                                                                                                                                                                                                                                                                         -1/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)),                                                                                                 cos(\\beta_{(5,8)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(5,8)} - \\phi_{(5,8)})**2*tan(\\beta_{(5,8)})**2)),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                     0,                                                                                                                  0,                                    0,                                                                               sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                                    0,                                                                             -sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}),                                                                                                                                                                                                cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                               sin(\\beta_{(7,8)})*cos(\\phi_{(7,8)}),                                                                                                                                                                                              -cos(\\delta_{(7,8)})/(sqrt(1 + 1/(cos(\\delta_{(7,8)} - \\phi_{(7,8)})**2*tan(\\beta_{(7,8)})**2))*cos(\\delta_{(7,8)} - \\phi_{(7,8)})*tan(\\beta_{(7,8)}))],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                     0,                                                                                                                  0,                                    0,                                                                               sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                                    0,                                                                             -sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}),                                                                                                                                                                                                sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                               sin(\\beta_{(7,8)})*sin(\\phi_{(7,8)}),                                                                                                                                                                                              -sin(\\delta_{(7,8)})/(sqrt(1 + 1/(cos(\\delta_{(7,8)} - \\phi_{(7,8)})**2*tan(\\beta_{(7,8)})**2))*cos(\\delta_{(7,8)} - \\phi_{(7,8)})*tan(\\beta_{(7,8)}))],\n[                                                                                              -cos(\\beta_{(1,2)}),                                     0,                                                                                                                  0,                                    0,                                                                                                 cos(\\beta_{(1,6)}),                                    0,                                                                                               -cos(\\beta_{(3,8)}),                                                                                                                                                                                                                                                                         -1/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)),                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                 cos(\\beta_{(7,8)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(7,8)} - \\phi_{(7,8)})**2*tan(\\beta_{(7,8)})**2))]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[-0.7190197069772716, 0.46653033334484806, 0.22545603076083248], [-0.1440711618041754, 0.08798224072063174, 0.22340958397007107], [-0.38075888779617006, 0.8092700976887757, 0.4348734722293312], [0.5762145744672778, 0.05391868826830025, 0.28012874635375756], [0.22053889329904677, 0.05755937143773849, 0.08499316673052616], [0.18531827430663095, 0.5706735108896537, 0.04336172002432798]])).all()


def test_spatial_R2SR():
    A = numpy.array(
        [[9, 1, 0, 1],
         [1, 9, 4, 0],
         [0, 0, 9, 4],
         [0, 0, 0, 9]]
    )
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[0],\n[0],\n[0],\n[0],\n[0],\n[0]]), Matrix([\n[                  -(a_y - r_{(1,4)y})*cos(\\beta_{(1,4)}) + (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}), 0, 0, 0, 0, 0, 0],\n[                   (a_x - r_{(1,4)x})*cos(\\beta_{(1,4)}) - (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}), 0, 0, 0, 0, 0, 0],\n[-(a_x - r_{(1,4)x})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}) + (a_y - r_{(1,4)y})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}), 0, 0, 0, 0, 0, 0],\n[                                                                              sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}), 0, 0, 0, 0, 0, 0],\n[                                                                              sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}), 0, 0, 0, 0, 0, 0],\n[                                                                                                cos(\\beta_{(1,4)}), 0, 0, 0, 0, 0, 0]]), Matrix([\n[                                                                              -(a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)})],\n[                                                                               (a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)})],\n[                                                            -(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)})],\n[                                                                                                                                          sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)})],\n[                                                                                                                                          sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)})],\n[                                                                                                                                                            cos(\\beta_{(1,2)})],\n[(r_{(2,3)x} - r_{(3,4)x})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}) + (r_{(2,3)y} - r_{(3,4)y})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (r_{(2,3)z} - r_{(3,4)z})*cos(\\beta_{(1,2)})]]), Matrix([\n[                  (a_y - r_{(1,4)y})*cos(\\beta_{(1,4)}) - (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                       0,        a_z - r_{(2,3)z},       -a_y + r_{(2,3)y},                 0,  a_z - r_{(3,4)z}, -a_y + r_{(3,4)y}],\n[                 -(a_x - r_{(1,4)x})*cos(\\beta_{(1,4)}) + (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),       -a_z + r_{(2,3)z},                       0,        a_x - r_{(2,3)x}, -a_z + r_{(3,4)z},                 0,  a_x - r_{(3,4)x}],\n[(a_x - r_{(1,4)x})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}) - (a_y - r_{(1,4)y})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),        a_y - r_{(2,3)y},       -a_x + r_{(2,3)x},                       0,  a_y - r_{(3,4)y}, -a_x + r_{(3,4)x},                 0],\n[                                                                            -sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                       1,                       0,                       0,                 1,                 0,                 0],\n[                                                                            -sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                       0,                       1,                       0,                 0,                 1,                 0],\n[                                                                                              -cos(\\beta_{(1,4)}),                       0,                       0,                       1,                 0,                 0,                 1],\n[                                                                                                                0, r_{(2,3)x} - r_{(3,4)x}, r_{(2,3)y} - r_{(3,4)y}, r_{(2,3)z} - r_{(3,4)z},                 0,                 0,                 0]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[-0.02032654749573748], [0.013547189835920806], [0.01984031747199766], [0.0626341203282414], [0.02526715383455612], [0.04691641116173456]])).all()


def test_spatial_3PHS():
    A = numpy.array(
        [[9, 2, 0, 2, 0, 2, 0, 0],
         [1, 9, 6, 0, 0, 0, 0, 0],
         [0, 0, 9, 0, 0, 0, 0, 4],
         [1, 0, 0, 9, 6, 0, 0, 0],
         [0, 0, 0, 0, 9, 0, 0, 4],
         [1, 0, 0, 0, 0, 9, 6, 0],
         [0, 0, 0, 0, 0, 0, 9, 4],
         [0, 0, 0, 0, 0, 0, 0, 9]]
       )
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0],\n[sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), 0, 0],\n[                  cos(\\beta_{(1,2)}), 0, 0],\n[                                   0, 0, 0],\n[                                   0, 0, 0],\n[                                   0, 0, 0]]), Matrix([\n[p_{(2,3)}*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)})/(2*pi) - (a_y - r_{(2,3)y})*cos(\\beta_{(2,3)}) + (a_z - r_{(2,3)z})*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                 0,  a_z - r_{(3,8)z}, -a_y + r_{(3,8)y}, 0, 0, 0, 0, 0, 0, 0, 0],\n[p_{(2,3)}*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)})/(2*pi) + (a_x - r_{(2,3)x})*cos(\\beta_{(2,3)}) - (a_z - r_{(2,3)z})*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}), -a_z + r_{(3,8)z},                 0,  a_x - r_{(3,8)x}, 0, 0, 0, 0, 0, 0, 0, 0],\n[p_{(2,3)}*cos(\\beta_{(2,3)})/(2*pi) - (a_x - r_{(2,3)x})*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}) + (a_y - r_{(2,3)y})*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),  a_y - r_{(3,8)y}, -a_x + r_{(3,8)x},                 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                                                                                                   sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                 1,                 0,                 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                                                                                                   sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                 0,                 1,                 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                                                                                                                     cos(\\beta_{(2,3)}),                 0,                 0,                 1, 0, 0, 0, 0, 0, 0, 0, 0]]), Matrix([\n[-sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                    0],\n[-sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                                    0],\n[                  -cos(\\beta_{(1,2)}),                   cos(\\beta_{(1,4)}),                                    0],\n[-sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                    0, sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)})],\n[-sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                    0, sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)})],\n[                  -cos(\\beta_{(1,2)}),                                    0,                   cos(\\beta_{(1,6)})],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0]]), Matrix([\n[-p_{(2,3)}*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)})/(2*pi) + (a_y - r_{(2,3)y})*cos(\\beta_{(2,3)}) - (a_z - r_{(2,3)z})*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                 0, -a_z + r_{(3,8)z},  a_y - r_{(3,8)y}, p_{(4,5)}*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)})/(2*pi) - (a_y - r_{(4,5)y})*cos(\\beta_{(4,5)}) + (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                 0,  a_z - r_{(5,8)z}, -a_y + r_{(5,8)y},                                                                                                                                                       0,                 0,                 0,                 0],\n[-p_{(2,3)}*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)})/(2*pi) - (a_x - r_{(2,3)x})*cos(\\beta_{(2,3)}) + (a_z - r_{(2,3)z})*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),  a_z - r_{(3,8)z},                 0, -a_x + r_{(3,8)x}, p_{(4,5)}*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)})/(2*pi) + (a_x - r_{(4,5)x})*cos(\\beta_{(4,5)}) - (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}), -a_z + r_{(5,8)z},                 0,  a_x - r_{(5,8)x},                                                                                                                                                       0,                 0,                 0,                 0],\n[-p_{(2,3)}*cos(\\beta_{(2,3)})/(2*pi) + (a_x - r_{(2,3)x})*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}) - (a_y - r_{(2,3)y})*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}), -a_y + r_{(3,8)y},  a_x - r_{(3,8)x},                 0, p_{(4,5)}*cos(\\beta_{(4,5)})/(2*pi) - (a_x - r_{(4,5)x})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}) + (a_y - r_{(4,5)y})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),  a_y - r_{(5,8)y}, -a_x + r_{(5,8)x},                 0,                                                                                                                                                       0,                 0,                 0,                 0],\n[-p_{(2,3)}*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)})/(2*pi) + (a_y - r_{(2,3)y})*cos(\\beta_{(2,3)}) - (a_z - r_{(2,3)z})*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                 0, -a_z + r_{(3,8)z},  a_y - r_{(3,8)y},                                                                                                                                                       0,                 0,                 0,                 0, p_{(6,7)}*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)})/(2*pi) - (a_y - r_{(6,7)y})*cos(\\beta_{(6,7)}) + (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                 0,  a_z - r_{(7,8)z}, -a_y + r_{(7,8)y}],\n[-p_{(2,3)}*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)})/(2*pi) - (a_x - r_{(2,3)x})*cos(\\beta_{(2,3)}) + (a_z - r_{(2,3)z})*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),  a_z - r_{(3,8)z},                 0, -a_x + r_{(3,8)x},                                                                                                                                                       0,                 0,                 0,                 0, p_{(6,7)}*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)})/(2*pi) + (a_x - r_{(6,7)x})*cos(\\beta_{(6,7)}) - (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}), -a_z + r_{(7,8)z},                 0,  a_x - r_{(7,8)x}],\n[-p_{(2,3)}*cos(\\beta_{(2,3)})/(2*pi) + (a_x - r_{(2,3)x})*sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}) - (a_y - r_{(2,3)y})*sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}), -a_y + r_{(3,8)y},  a_x - r_{(3,8)x},                 0,                                                                                                                                                       0,                 0,                 0,                 0, p_{(6,7)}*cos(\\beta_{(6,7)})/(2*pi) - (a_x - r_{(6,7)x})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}) + (a_y - r_{(6,7)y})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),  a_y - r_{(7,8)y}, -a_x + r_{(7,8)x},                 0],\n[                                                                                                                   -sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                -1,                 0,                 0,                                                                                                                    sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                 1,                 0,                 0,                                                                                                                                                       0,                 0,                 0,                 0],\n[                                                                                                                   -sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                 0,                -1,                 0,                                                                                                                    sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                 0,                 1,                 0,                                                                                                                                                       0,                 0,                 0,                 0],\n[                                                                                                                                     -cos(\\beta_{(2,3)}),                 0,                 0,                -1,                                                                                                                                      cos(\\beta_{(4,5)}),                 0,                 0,                 1,                                                                                                                                                       0,                 0,                 0,                 0],\n[                                                                                                                   -sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                -1,                 0,                 0,                                                                                                                                                       0,                 0,                 0,                 0,                                                                                                                    sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                 1,                 0,                 0],\n[                                                                                                                   -sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),                 0,                -1,                 0,                                                                                                                                                       0,                 0,                 0,                 0,                                                                                                                    sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                 0,                 1,                 0],\n[                                                                                                                                     -cos(\\beta_{(2,3)}),                 0,                 0,                -1,                                                                                                                                                       0,                 0,                 0,                 0,                                                                                                                                      cos(\\beta_{(6,7)}),                 0,                 0,                 1]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[0.5908830416374078, -0.06515710297640874, -0.30754272049567766],
    [-0.9459776844232199, 0.5758233704071427, 0.884748496120456],
    [0.46623204082803915, 0.24040401251881377, 0.1262974111351349],
    [1.494197522406394, -1.2473704797886107, 0.06730095220669222],
    [-0.2193355597507754, -0.6889826191018327, 0.7725465269574716],
    [-0.04373923749924692, 1.2337412328299768, -0.5923105366105943]])).all()



def test_spatial_3RFS():
    A = numpy.array(
        [[9, 1, 0, 1, 0, 1, 0, 0],
         [1, 9, 7, 0, 0, 0, 0, 0],
         [0, 0, 9, 0, 0, 0, 0, 4],
         [1, 0, 0, 9, 7, 0, 0, 0],
         [0, 0, 0, 0, 9, 0, 0, 4],
         [1, 0, 0, 0, 0, 9, 7, 0],
         [0, 0, 0, 0, 0, 0, 9, 4],
         [0, 0, 0, 0, 0, 0, 0, 9]]
    )
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[                  -(a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), 0, 0],\n[                   (a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0],\n[-(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0],\n[                                                                              sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0],\n[                                                                              sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), 0, 0],\n[                                                                                                cos(\\beta_{(1,2)}), 0, 0]]), Matrix([\n[sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                 0,  a_z - r_{(3,8)z}, -a_y + r_{(3,8)y}, 0, 0, 0, 0, 0, 0, 0, 0],\n[sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}), -a_z + r_{(3,8)z},                 0,  a_x - r_{(3,8)x}, 0, 0, 0, 0, 0, 0, 0, 0],\n[                  cos(\\beta_{(2,3)}),  a_y - r_{(3,8)y}, -a_x + r_{(3,8)x},                 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                   0,                 1,                 0,                 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                   0,                 0,                 1,                 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                   0,                 0,                 0,                 1, 0, 0, 0, 0, 0, 0, 0, 0]]), Matrix([\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                   -(a_y - r_{(1,4)y})*cos(\\beta_{(1,4)}) + (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                                                                                                                  0],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                    (a_x - r_{(1,4)x})*cos(\\beta_{(1,4)}) - (a_z - r_{(1,4)z})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                                                                                                  0],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -(a_x - r_{(1,4)x})*sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}) + (a_y - r_{(1,4)y})*sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                                                                                                  0],\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                  0,                   -(a_y - r_{(1,6)y})*cos(\\beta_{(1,6)}) + (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)})],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                  0,                    (a_x - r_{(1,6)x})*cos(\\beta_{(1,6)}) - (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)})],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                  0, -(a_x - r_{(1,6)x})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}) + (a_y - r_{(1,6)y})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)})],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                               sin(\\beta_{(1,4)})*cos(\\phi_{(1,4)}),                                                                                                                  0],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                               sin(\\beta_{(1,4)})*sin(\\phi_{(1,4)}),                                                                                                                  0],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                 cos(\\beta_{(1,4)}),                                                                                                                  0],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                  0,                                                                               sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)})],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                  0,                                                                               sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)})],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                                  0,                                                                                                 cos(\\beta_{(1,6)})]]), Matrix([\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                 0, -a_z + r_{(3,8)z},  a_y - r_{(3,8)y}, sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                 0,  a_z - r_{(5,8)z}, -a_y + r_{(5,8)y},                                    0,                 0,                 0,                 0],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),  a_z - r_{(3,8)z},                 0, -a_x + r_{(3,8)x}, sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}), -a_z + r_{(5,8)z},                 0,  a_x - r_{(5,8)x},                                    0,                 0,                 0,                 0],\n[                  -cos(\\beta_{(2,3)}), -a_y + r_{(3,8)y},  a_x - r_{(3,8)x},                 0,                   cos(\\beta_{(4,5)}),  a_y - r_{(5,8)y}, -a_x + r_{(5,8)x},                 0,                                    0,                 0,                 0,                 0],\n[-sin(\\beta_{(2,3)})*cos(\\phi_{(2,3)}),                 0, -a_z + r_{(3,8)z},  a_y - r_{(3,8)y},                                    0,                 0,                 0,                 0, sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                 0,  a_z - r_{(7,8)z}, -a_y + r_{(7,8)y}],\n[-sin(\\beta_{(2,3)})*sin(\\phi_{(2,3)}),  a_z - r_{(3,8)z},                 0, -a_x + r_{(3,8)x},                                    0,                 0,                 0,                 0, sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}), -a_z + r_{(7,8)z},                 0,  a_x - r_{(7,8)x}],\n[                  -cos(\\beta_{(2,3)}), -a_y + r_{(3,8)y},  a_x - r_{(3,8)x},                 0,                                    0,                 0,                 0,                 0,                   cos(\\beta_{(6,7)}),  a_y - r_{(7,8)y}, -a_x + r_{(7,8)x},                 0],\n[                                    0,                -1,                 0,                 0,                                    0,                 1,                 0,                 0,                                    0,                 0,                 0,                 0],\n[                                    0,                 0,                -1,                 0,                                    0,                 0,                 1,                 0,                                    0,                 0,                 0,                 0],\n[                                    0,                 0,                 0,                -1,                                    0,                 0,                 0,                 1,                                    0,                 0,                 0,                 0],\n[                                    0,                -1,                 0,                 0,                                    0,                 0,                 0,                 0,                                    0,                 1,                 0,                 0],\n[                                    0,                 0,                -1,                 0,                                    0,                 0,                 0,                 0,                                    0,                 0,                 1,                 0],\n[                                    0,                 0,                 0,                -1,                                    0,                 0,                 0,                 0,                                    0,                 0,                 0,                 1]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[-0.03626461, -0.43059102,  0.14049168],
       [ 0.03859895, -0.07875818,  0.07839828],
       [-0.08266009, -0.42143952,  0.28195617],
       [-0.27626281, -0.22535705,  0.499094  ],
       [ 0.11695045,  1.35715678, -0.34503766],
       [ 0.01347087,  0.06018991,  0.196297  ]])).all()


def test_planar_for_descending():
    A = numpy.array(
      [[9, 2, 0, 0, 2, 0, 0],
       [1, 9, 0, 2, 0, 0, 0],
       [0, 0, 9, 1, 0, 0, 1],
       [0, 0, 0, 9, 2, 1, 0],
       [1, 0, 0, 0, 9, 0, 0],
       [0, 0, 0, 1, 0, 9, 1],
       [0, 0, 0, 0, 0, 0, 9]]
    )
    jac = Jacobian(A, robot_type = 'planar')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[cos(\\phi_{(1,2)}), 0, 0],\n[sin(\\phi_{(1,2)}), 0, 0],\n[                0, 0, 0]]), Matrix([\n[cos(\\phi_{(2,4)}),  a_y - r_{(3,4)y}, -a_y + r_{(3,7)y}, 0, 0],\n[sin(\\phi_{(2,4)}), -a_x + r_{(3,4)x},  a_x - r_{(3,7)x}, 0, 0],\n[                0,                -1,                 1, 0, 0]]), Matrix([\n[                 0,                 0, -a_y + r_{(4,6)y}],\n[                 0,                 0,  a_x - r_{(4,6)x}],\n[-cos(\\phi_{(1,2)}), cos(\\phi_{(1,5)}),                 0],\n[-sin(\\phi_{(1,2)}), sin(\\phi_{(1,5)}),                 0],\n[                 0,                 0,                 1]]), Matrix([\n[                 0, -a_y + r_{(3,4)y},  a_y - r_{(3,7)y},                  0, -a_y + r_{(6,7)y}],\n[                 0,  a_x - r_{(3,4)x}, -a_x + r_{(3,7)x},                  0,  a_x - r_{(6,7)x}],\n[-cos(\\phi_{(2,4)}),                 0,                 0, -cos(\\phi_{(4,5)}),                 0],\n[-sin(\\phi_{(2,4)}),                 0,                 0, -sin(\\phi_{(4,5)}),                 0],\n[                 0,                 1,                -1,                  0,                 1]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(2)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[-0.11882638,  0.50774852, -0.58812428],
       [-0.14723595,  0.34943317, -0.74914152],
       [ 0.        ,  0.        ,  2.37618934]])).all()

def test_spatial_for_descending_1():
    A = numpy.array(
      [[9, 0, 0, 0, 1, 0, 0, 0, 1, 0, 0, 0, 0, 0, 1, 0],
       [0, 9, 0, 0, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0, 4],
       [0, 0, 9, 0, 0, 0, 4, 0, 0, 0, 0, 0, 0, 0, 2, 0],
       [0, 0, 0, 9, 0, 0, 4, 2, 0, 0, 0, 0, 0, 0, 0, 0],
       [0, 0, 0, 0, 9, 0, 0, 0, 0, 0, 0, 2, 0, 0, 0, 0],
       [0, 0, 0, 0, 0, 9, 5, 0, 0, 0, 0, 0, 0, 0, 0, 2],
       [0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 4, 4, 4, 4, 0, 0],
       [0, 0, 0, 1, 0, 0, 0, 9, 0, 0, 0, 0, 0, 0, 0, 4],
       [0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 0, 2, 0, 0],
       [0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 2, 0, 0, 4],
       [0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 0, 0],
       [0, 0, 0, 0, 1, 0, 0, 0, 0, 0, 0, 9, 0, 0, 0, 0],
       [0, 0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 9, 0, 0, 0],
       [0, 0, 0, 0, 0, 0, 0, 0, 1, 0, 0, 0, 0, 9, 0, 0],
       [0, 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9, 0],
       [0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 9]])
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[0, 0, 0, sin(\\beta_{(5,12)})*cos(\\phi_{(5,12)}), 0, 0],\n[0, 0, 0, sin(\\beta_{(5,12)})*sin(\\phi_{(5,12)}), 0, 0],\n[0, 0, 0,                    cos(\\beta_{(5,12)}), 0, 0],\n[0, 0, 0,                                      0, 0, 0],\n[0, 0, 0,                                      0, 0, 0],\n[0, 0, 0,                                      0, 0, 0]]), Matrix([\n[                  -(a_y - r_{(1,5)y})*cos(\\beta_{(1,5)}) + (a_z - r_{(1,5)z})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                   (a_y - r_{(6,7)y})*cos(\\beta_{(6,7)}) - (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                                                                              (a_y - r_{(6,7)y})/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)) + (a_z - r_{(6,7)z})*sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})), sin(\\beta_{(6,16)})*cos(\\phi_{(6,16)}), 0, 0, 0,                  0, -a_z + r_{(7,12)z},  a_y - r_{(7,12)y}, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                   (a_x - r_{(1,5)x})*cos(\\beta_{(1,5)}) - (a_z - r_{(1,5)z})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                  -(a_x - r_{(6,7)x})*cos(\\beta_{(6,7)}) + (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                                                                             -(a_x - r_{(6,7)x})/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)) - (a_z - r_{(6,7)z})*cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})), sin(\\beta_{(6,16)})*sin(\\phi_{(6,16)}), 0, 0, 0,  a_z - r_{(7,12)z},                  0, -a_x + r_{(7,12)x}, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[-(a_x - r_{(1,5)x})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}) + (a_y - r_{(1,5)y})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, (a_x - r_{(6,7)x})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}) - (a_y - r_{(6,7)y})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}), -(a_x - r_{(6,7)x})*sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})) + (a_y - r_{(6,7)y})*cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                    cos(\\beta_{(6,16)}), 0, 0, 0, -a_y + r_{(7,12)y},  a_x - r_{(7,12)x},                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                                                              sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                                                                             -sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                                                                                                                                                                                                cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                                      0, 0, 0, 0,                 -1,                  0,                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                                                              sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                                                                             -sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                                                                                                                                                                                                sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                                      0, 0, 0, 0,                  0,                 -1,                  0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0],\n[                                                                                                cos(\\beta_{(1,5)}), 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0,                                                                                               -cos(\\beta_{(6,7)}),                                                                                                                                                                                                                                                                         -1/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)),                                      0, 0, 0, 0,                  0,                  0,                 -1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0]]), Matrix([\n[                                      0,                                       0,                                    0, -sin(\\beta_{(5,12)})*cos(\\phi_{(5,12)}), sin(\\beta_{(9,14)})*cos(\\phi_{(9,14)}),                                         0],\n[                                      0,                                       0,                                    0, -sin(\\beta_{(5,12)})*sin(\\phi_{(5,12)}), sin(\\beta_{(9,14)})*sin(\\phi_{(9,14)}),                                         0],\n[                                      0,                                       0,                                    0,                    -cos(\\beta_{(5,12)}),                    cos(\\beta_{(9,14)}),                                         0],\n[                                      0, -sin(\\beta_{(3,15)})*cos(\\phi_{(3,15)}),                                    0, -sin(\\beta_{(5,12)})*cos(\\phi_{(5,12)}),                                      0,                                         0],\n[                                      0, -sin(\\beta_{(3,15)})*sin(\\phi_{(3,15)}),                                    0, -sin(\\beta_{(5,12)})*sin(\\phi_{(5,12)}),                                      0,                                         0],\n[                                      0,                    -cos(\\beta_{(3,15)}),                                    0,                    -cos(\\beta_{(5,12)}),                                      0,                                         0],\n[                                      0,                                       0, sin(\\beta_{(4,8)})*cos(\\phi_{(4,8)}),                                       0,                                      0,                                         0],\n[                                      0,                                       0, sin(\\beta_{(4,8)})*sin(\\phi_{(4,8)}),                                       0,                                      0,                                         0],\n[                                      0,                                       0,                   cos(\\beta_{(4,8)}),                                       0,                                      0,                                         0],\n[-sin(\\beta_{(2,11)})*cos(\\phi_{(2,11)}),                                       0,                                    0,                                       0,                                      0,                                         0],\n[-sin(\\beta_{(2,11)})*sin(\\phi_{(2,11)}),                                       0,                                    0,                                       0,                                      0,                                         0],\n[                   -cos(\\beta_{(2,11)}),                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0, -sin(\\beta_{(10,13)})*cos(\\phi_{(10,13)})],\n[                                      0,                                       0,                                    0,                                       0,                                      0, -sin(\\beta_{(10,13)})*sin(\\phi_{(10,13)})],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                     -cos(\\beta_{(10,13)})],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0],\n[                                      0,                                       0,                                    0,                                       0,                                      0,                                         0]]), Matrix([\n[                                                                                        (a_y - r_{(1,5)y})*cos(\\beta_{(1,5)}) - (a_z - r_{(1,5)z})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}),                   -(a_y - r_{(1,9)y})*cos(\\beta_{(1,9)}) + (a_z - r_{(1,9)z})*sin(\\beta_{(1,9)})*sin(\\phi_{(1,9)}),                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,                           0,           a_z - r_{(7,12)z},          -a_y + r_{(7,12)y},                          0,                          0,                          0,                  0, -a_z + r_{(7,14)z},  a_y - r_{(7,14)y},                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                       -(a_x - r_{(1,5)x})*cos(\\beta_{(1,5)}) + (a_z - r_{(1,5)z})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}),                    (a_x - r_{(1,9)x})*cos(\\beta_{(1,9)}) - (a_z - r_{(1,9)z})*sin(\\beta_{(1,9)})*cos(\\phi_{(1,9)}),                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,          -a_z + r_{(7,12)z},                           0,           a_x - r_{(7,12)x},                          0,                          0,                          0,  a_z - r_{(7,14)z},                  0, -a_x + r_{(7,14)x},                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                      (a_x - r_{(1,5)x})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}) - (a_y - r_{(1,5)y})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}), -(a_x - r_{(1,9)x})*sin(\\beta_{(1,9)})*sin(\\phi_{(1,9)}) + (a_y - r_{(1,9)y})*sin(\\beta_{(1,9)})*cos(\\phi_{(1,9)}),                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,           a_y - r_{(7,12)y},          -a_x + r_{(7,12)x},                           0,                          0,                          0,                          0, -a_y + r_{(7,14)y},  a_x - r_{(7,14)x},                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                        (a_y - r_{(1,5)y})*cos(\\beta_{(1,5)}) - (a_z - r_{(1,5)z})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}),                                                                                                                  0,                    -(a_y - r_{(1,15)y})*cos(\\beta_{(1,15)}) + (a_z - r_{(1,15)z})*sin(\\beta_{(1,15)})*sin(\\phi_{(1,15)}),                  0,                  0,                  0,                 0,  a_z - r_{(3,7)z}, -a_y + r_{(3,7)y},                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,                           0,           a_z - r_{(7,12)z},          -a_y + r_{(7,12)y},                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                       -(a_x - r_{(1,5)x})*cos(\\beta_{(1,5)}) + (a_z - r_{(1,5)z})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}),                                                                                                                  0,                     (a_x - r_{(1,15)x})*cos(\\beta_{(1,15)}) - (a_z - r_{(1,15)z})*sin(\\beta_{(1,15)})*cos(\\phi_{(1,15)}),                  0,                  0,                  0, -a_z + r_{(3,7)z},                 0,  a_x - r_{(3,7)x},                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,          -a_z + r_{(7,12)z},                           0,           a_x - r_{(7,12)x},                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                      (a_x - r_{(1,5)x})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}) - (a_y - r_{(1,5)y})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}),                                                                                                                  0, -(a_x - r_{(1,15)x})*sin(\\beta_{(1,15)})*sin(\\phi_{(1,15)}) + (a_y - r_{(1,15)y})*sin(\\beta_{(1,15)})*cos(\\phi_{(1,15)}),                  0,                  0,                  0,  a_y - r_{(3,7)y}, -a_x + r_{(3,7)x},                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,           a_y - r_{(7,12)y},          -a_x + r_{(7,12)x},                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,         -a_z + r_{(4,7)z},          a_y - r_{(4,7)y},                   -(a_y - r_{(6,7)y})*cos(\\beta_{(6,7)}) + (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                                                                            -(a_y - r_{(6,7)y})/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)) - (a_z - r_{(6,7)z})*sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})), -sin(\\beta_{(6,16)})*cos(\\phi_{(6,16)}),                         0,                         0,                         0,                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,  a_z - r_{(8,16)z}, -a_y + r_{(8,16)y},                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,          a_z - r_{(4,7)z},                         0,         -a_x + r_{(4,7)x},                    (a_x - r_{(6,7)x})*cos(\\beta_{(6,7)}) - (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                                                                             (a_x - r_{(6,7)x})/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)) + (a_z - r_{(6,7)z})*cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})), -sin(\\beta_{(6,16)})*sin(\\phi_{(6,16)}),                         0,                         0,                         0,                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0, -a_z + r_{(8,16)z},                  0,  a_x - r_{(8,16)x},                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,         -a_y + r_{(4,7)y},          a_x - r_{(4,7)x},                         0, -(a_x - r_{(6,7)x})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}) + (a_y - r_{(6,7)y})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}), (a_x - r_{(6,7)x})*sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})) - (a_y - r_{(6,7)y})*cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                    -cos(\\beta_{(6,16)}),                         0,                         0,                         0,                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,  a_y - r_{(8,16)y}, -a_x + r_{(8,16)x},                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,  a_z - r_{(2,16)z}, -a_y + r_{(2,16)y},                 0,                 0,                 0,                         0,                         0,                         0,                   -(a_y - r_{(6,7)y})*cos(\\beta_{(6,7)}) + (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                                                                            -(a_y - r_{(6,7)y})/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)) - (a_z - r_{(6,7)z})*sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})), -sin(\\beta_{(6,16)})*cos(\\phi_{(6,16)}),                         0,         a_z - r_{(7,11)z},        -a_y + r_{(7,11)y},                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0, -a_z + r_{(2,16)z},                  0,  a_x - r_{(2,16)x},                 0,                 0,                 0,                         0,                         0,                         0,                    (a_x - r_{(6,7)x})*cos(\\beta_{(6,7)}) - (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                                                                             (a_x - r_{(6,7)x})/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)) + (a_z - r_{(6,7)z})*cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})), -sin(\\beta_{(6,16)})*sin(\\phi_{(6,16)}),        -a_z + r_{(7,11)z},                         0,         a_x - r_{(7,11)x},                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,  a_y - r_{(2,16)y}, -a_x + r_{(2,16)x},                  0,                 0,                 0,                 0,                         0,                         0,                         0, -(a_x - r_{(6,7)x})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}) + (a_y - r_{(6,7)y})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}), (a_x - r_{(6,7)x})*sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})) - (a_y - r_{(6,7)y})*cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                    -cos(\\beta_{(6,16)}),         a_y - r_{(7,11)y},        -a_x + r_{(7,11)x},                         0,                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                   -(a_y - r_{(6,7)y})*cos(\\beta_{(6,7)}) + (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                                                                            -(a_y - r_{(6,7)y})/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)) - (a_z - r_{(6,7)z})*sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})), -sin(\\beta_{(6,16)})*cos(\\phi_{(6,16)}),                         0,                         0,                         0,                           0,                           0,                           0,                          0,          a_z - r_{(7,13)z},         -a_y + r_{(7,13)y},                  0,                  0,                  0,                  0,                  0,                  0,                   0,  a_z - r_{(10,16)z}, -a_y + r_{(10,16)y}],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                    (a_x - r_{(6,7)x})*cos(\\beta_{(6,7)}) - (a_z - r_{(6,7)z})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                                                                             (a_x - r_{(6,7)x})/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)) + (a_z - r_{(6,7)z})*cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})), -sin(\\beta_{(6,16)})*sin(\\phi_{(6,16)}),                         0,                         0,                         0,                           0,                           0,                           0,         -a_z + r_{(7,13)z},                          0,          a_x - r_{(7,13)x},                  0,                  0,                  0,                  0,                  0,                  0, -a_z + r_{(10,16)z},                   0,  a_x - r_{(10,16)x}],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0, -(a_x - r_{(6,7)x})*sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}) + (a_y - r_{(6,7)y})*sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}), (a_x - r_{(6,7)x})*sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})) - (a_y - r_{(6,7)y})*cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                    -cos(\\beta_{(6,16)}),                         0,                         0,                         0,                           0,                           0,                           0,          a_y - r_{(7,13)y},         -a_x + r_{(7,13)x},                          0,                  0,                  0,                  0,                  0,                  0,                  0,  a_y - r_{(10,16)y}, -a_x + r_{(10,16)x},                   0],\n[                                                                                                                                                  -sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}),                                                                               sin(\\beta_{(1,9)})*cos(\\phi_{(1,9)}),                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,                           1,                           0,                           0,                          0,                          0,                          0,                 -1,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                  -sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}),                                                                               sin(\\beta_{(1,9)})*sin(\\phi_{(1,9)}),                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,                           0,                           1,                           0,                          0,                          0,                          0,                  0,                 -1,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                    -cos(\\beta_{(1,5)}),                                                                                                 cos(\\beta_{(1,9)}),                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,                           0,                           0,                           1,                          0,                          0,                          0,                  0,                  0,                 -1,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                  -sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}),                                                                                                                  0,                                                                                   sin(\\beta_{(1,15)})*cos(\\phi_{(1,15)}),                  0,                  0,                  0,                 1,                 0,                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,                           1,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                  -sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}),                                                                                                                  0,                                                                                   sin(\\beta_{(1,15)})*sin(\\phi_{(1,15)}),                  0,                  0,                  0,                 0,                 1,                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,                           0,                           1,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                    -cos(\\beta_{(1,5)}),                                                                                                                  0,                                                                                                      cos(\\beta_{(1,15)}),                  0,                  0,                  0,                 0,                 0,                 1,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,                           0,                           0,                           1,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                        -1,                         0,                         0,                                                                               sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                                                                                                                                                                                              -cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                                       0,                         0,                         0,                         0,                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  1,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                        -1,                         0,                                                                               sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                                                                                                                                                                                              -sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                                       0,                         0,                         0,                         0,                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  1,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                        -1,                                                                                                 cos(\\beta_{(6,7)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)),                                       0,                         0,                         0,                         0,                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  1,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  1,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                               sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                                                                                                                                                                                              -cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                                       0,                         1,                         0,                         0,                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  1,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                               sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                                                                                                                                                                                              -sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                                       0,                         0,                         1,                         0,                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  1,                 0,                 0,                 0,                         0,                         0,                         0,                                                                                                 cos(\\beta_{(6,7)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)),                                       0,                         0,                         0,                         1,                           0,                           0,                           0,                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                               sin(\\beta_{(6,7)})*cos(\\phi_{(6,7)}),                                                                                                                                                                                              -cos(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                                       0,                         0,                         0,                         0,                           0,                           0,                           0,                          1,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   1,                   0,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                               sin(\\beta_{(6,7)})*sin(\\phi_{(6,7)}),                                                                                                                                                                                              -sin(\\delta_{(6,7)})/(sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2))*cos(\\delta_{(6,7)} - \\phi_{(6,7)})*tan(\\beta_{(6,7)})),                                       0,                         0,                         0,                         0,                           0,                           0,                           0,                          0,                          1,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   1,                   0],\n[                                                                                                                                                                                      0,                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                                                 cos(\\beta_{(6,7)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(6,7)} - \\phi_{(6,7)})**2*tan(\\beta_{(6,7)})**2)),                                       0,                         0,                         0,                         0,                           0,                           0,                           0,                          0,                          0,                          1,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   1],\n[   (r_{(2,16)x} - r_{(7,11)x})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}) + (r_{(2,16)y} - r_{(7,11)y})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}) + (r_{(2,16)z} - r_{(7,11)z})*cos(\\beta_{(1,5)}),                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0, r_{(2,16)x} - r_{(7,11)x}, r_{(2,16)y} - r_{(7,11)y}, r_{(2,16)z} - r_{(7,11)z},  -r_{(2,16)x} + r_{(7,11)x},  -r_{(2,16)y} + r_{(7,11)y},  -r_{(2,16)z} + r_{(7,11)z},                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[      (r_{(4,7)x} - r_{(8,16)x})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}) + (r_{(4,7)y} - r_{(8,16)y})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}) + (r_{(4,7)z} - r_{(8,16)z})*cos(\\beta_{(1,5)}),                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0, -r_{(4,7)x} + r_{(8,16)x}, -r_{(4,7)y} + r_{(8,16)y}, -r_{(4,7)z} + r_{(8,16)z},                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0,   -r_{(4,7)x} + r_{(8,16)x},   -r_{(4,7)y} + r_{(8,16)y},   -r_{(4,7)z} + r_{(8,16)z},                          0,                          0,                          0,                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0],\n[(r_{(10,16)x} - r_{(7,13)x})*sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}) + (r_{(10,16)y} - r_{(7,13)y})*sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}) + (r_{(10,16)z} - r_{(7,13)z})*cos(\\beta_{(1,5)}),                                                                                                                  0,                                                                                                                        0,                  0,                  0,                  0,                 0,                 0,                 0,                         0,                         0,                         0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                       0,                         0,                         0,                         0, -r_{(10,16)x} + r_{(7,13)x}, -r_{(10,16)y} + r_{(7,13)y}, -r_{(10,16)z} + r_{(7,13)z}, r_{(10,16)x} - r_{(7,13)x}, r_{(10,16)y} - r_{(7,13)y}, r_{(10,16)z} - r_{(7,13)z},                  0,                  0,                  0,                  0,                  0,                  0,                   0,                   0,                   0]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[ 1.55858149e+01, -3.03882002e-01, -2.40798360e+01,
         2.24362525e-01, -2.06690469e-01, -1.21121995e+02],
       [-1.13832218e+01,  1.77379830e+00,  1.72827684e+01,
        -8.27101773e-01,  1.89266970e+00,  8.59600651e+01],
       [ 2.67482562e+00, -9.22011936e-01, -4.42133618e+00,
         5.14047209e-01, -1.37583419e-01, -2.21719379e+01],
       [ 1.28301298e+01, -1.07326109e+01, -2.01307156e+01,
         5.73874124e+00, -1.13629612e+01, -9.88775596e+01],
       [ 9.22973346e+00, -5.91668151e-01, -1.44700515e+01,
         2.06073096e+00, -2.03709917e+00, -7.11180403e+01],
       [-5.74486260e+01, -2.71642977e+00,  8.60002598e+01,
         2.85012796e-01, -1.15294950e+00,  4.38284057e+02]])).all()


def test_spatial_for_descending_2():
    A = numpy.array(
      [[9, 0, 0, 0, 0, 1, 0],
       [0, 9, 0, 1, 0, 1, 0],
       [0, 0, 9, 1, 1, 0, 0],
       [0, 1, 1, 9, 0, 0, 0],
       [0, 0, 1, 0, 9, 0, 1],
       [1, 1, 0, 0, 0, 9, 0],
       [0, 0, 0, 0, 1, 0, 9]])
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[                  -(a_y - r_{(1,6)y})*cos(\\beta_{(1,6)}) + (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                   -(a_y - r_{(2,4)y})*cos(\\beta_{(2,4)}) + (a_z - r_{(2,4)z})*sin(\\beta_{(2,4)})*sin(\\phi_{(2,4)}),                   (a_y - r_{(2,6)y})*cos(\\beta_{(2,6)}) - (a_z - r_{(2,6)z})*sin(\\beta_{(2,6)})*sin(\\phi_{(2,6)}),                   (a_y - r_{(3,4)y})*cos(\\beta_{(3,4)}) - (a_z - r_{(3,4)z})*sin(\\beta_{(3,4)})*sin(\\phi_{(3,4)}),                   -(a_y - r_{(3,5)y})*cos(\\beta_{(3,5)}) + (a_z - r_{(3,5)z})*sin(\\beta_{(3,5)})*sin(\\phi_{(3,5)}),                   -(a_y - r_{(5,7)y})*cos(\\beta_{(5,7)}) + (a_z - r_{(5,7)z})*sin(\\beta_{(5,7)})*sin(\\phi_{(5,7)})],\n[                   (a_x - r_{(1,6)x})*cos(\\beta_{(1,6)}) - (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                    (a_x - r_{(2,4)x})*cos(\\beta_{(2,4)}) - (a_z - r_{(2,4)z})*sin(\\beta_{(2,4)})*cos(\\phi_{(2,4)}),                  -(a_x - r_{(2,6)x})*cos(\\beta_{(2,6)}) + (a_z - r_{(2,6)z})*sin(\\beta_{(2,6)})*cos(\\phi_{(2,6)}),                  -(a_x - r_{(3,4)x})*cos(\\beta_{(3,4)}) + (a_z - r_{(3,4)z})*sin(\\beta_{(3,4)})*cos(\\phi_{(3,4)}),                    (a_x - r_{(3,5)x})*cos(\\beta_{(3,5)}) - (a_z - r_{(3,5)z})*sin(\\beta_{(3,5)})*cos(\\phi_{(3,5)}),                    (a_x - r_{(5,7)x})*cos(\\beta_{(5,7)}) - (a_z - r_{(5,7)z})*sin(\\beta_{(5,7)})*cos(\\phi_{(5,7)})],\n[-(a_x - r_{(1,6)x})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}) + (a_y - r_{(1,6)y})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}), -(a_x - r_{(2,4)x})*sin(\\beta_{(2,4)})*sin(\\phi_{(2,4)}) + (a_y - r_{(2,4)y})*sin(\\beta_{(2,4)})*cos(\\phi_{(2,4)}), (a_x - r_{(2,6)x})*sin(\\beta_{(2,6)})*sin(\\phi_{(2,6)}) - (a_y - r_{(2,6)y})*sin(\\beta_{(2,6)})*cos(\\phi_{(2,6)}), (a_x - r_{(3,4)x})*sin(\\beta_{(3,4)})*sin(\\phi_{(3,4)}) - (a_y - r_{(3,4)y})*sin(\\beta_{(3,4)})*cos(\\phi_{(3,4)}), -(a_x - r_{(3,5)x})*sin(\\beta_{(3,5)})*sin(\\phi_{(3,5)}) + (a_y - r_{(3,5)y})*sin(\\beta_{(3,5)})*cos(\\phi_{(3,5)}), -(a_x - r_{(5,7)x})*sin(\\beta_{(5,7)})*sin(\\phi_{(5,7)}) + (a_y - r_{(5,7)y})*sin(\\beta_{(5,7)})*cos(\\phi_{(5,7)})],\n[                                                                              sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                                                                               sin(\\beta_{(2,4)})*cos(\\phi_{(2,4)}),                                                                             -sin(\\beta_{(2,6)})*cos(\\phi_{(2,6)}),                                                                             -sin(\\beta_{(3,4)})*cos(\\phi_{(3,4)}),                                                                               sin(\\beta_{(3,5)})*cos(\\phi_{(3,5)}),                                                                               sin(\\beta_{(5,7)})*cos(\\phi_{(5,7)})],\n[                                                                              sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                                                                               sin(\\beta_{(2,4)})*sin(\\phi_{(2,4)}),                                                                             -sin(\\beta_{(2,6)})*sin(\\phi_{(2,6)}),                                                                             -sin(\\beta_{(3,4)})*sin(\\phi_{(3,4)}),                                                                               sin(\\beta_{(3,5)})*sin(\\phi_{(3,5)}),                                                                               sin(\\beta_{(5,7)})*sin(\\phi_{(5,7)})],\n[                                                                                                cos(\\beta_{(1,6)}),                                                                                                 cos(\\beta_{(2,4)}),                                                                                               -cos(\\beta_{(2,6)}),                                                                                               -cos(\\beta_{(3,4)}),                                                                                                 cos(\\beta_{(3,5)}),                                                                                                 cos(\\beta_{(5,7)})]]), None, None, None]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[ 0.        , -0.24717605,  0.1100608 ,  0.61829149,  0.07139447,
        -0.56438892],
       [ 0.        ,  0.16473732,  0.17374646, -0.43285794, -0.40411481,
         0.0770093 ],
       [ 0.        ,  0.24126337, -0.16224528, -0.06686038, -0.02298567,
         0.27919047],
       [ 0.47249501,  0.76164703, -0.79696467, -0.45264098,  0.69873959,
         0.45623812],
       [ 0.21307796,  0.3072551 , -0.05670865, -0.53653983,  0.08303055,
         0.20082922],
       [ 0.85518784,  0.57051564, -0.601358  , -0.71220865,  0.71054128,
         0.8668993 ]])).all()

def test_spatial_for_descending_3():
    A = numpy.array(
      [[9, 0, 0, 0, 2, 2, 2, 0],
       [0, 9, 0, 0, 0, 0, 3, 5],
       [0, 0, 9, 0, 0, 3, 0, 5],
       [0, 0, 0, 9, 3, 0, 0, 5],
       [1, 0, 0, 0, 9, 0, 0, 0],
       [1, 0, 0, 0, 0, 9, 0, 0],
       [1, 0, 0, 0, 0, 0, 9, 0],
       [0, 0, 0, 0, 0, 0, 0, 9]])
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}), 0, 0],\n[sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}), 0, 0],\n[                  cos(\\beta_{(1,5)}), 0, 0],\n[                                   0, 0, 0],\n[                                   0, 0, 0],\n[                                   0, 0, 0]]), Matrix([\n[0, 0, 0, 0, 0, 0, 0, 0,                   (a_y - r_{(4,5)y})*cos(\\beta_{(4,5)}) - (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}), -sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                   -(a_y - r_{(4,8)y})*cos(\\beta_{(4,8)}) + (a_z - r_{(4,8)z})*sin(\\beta_{(4,8)})*sin(\\phi_{(4,8)}),                                                                            -(a_y - r_{(4,8)y})/sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2)) - (a_z - r_{(4,8)z})*sin(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[0, 0, 0, 0, 0, 0, 0, 0,                  -(a_x - r_{(4,5)x})*cos(\\beta_{(4,5)}) + (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}), -sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                    (a_x - r_{(4,8)x})*cos(\\beta_{(4,8)}) - (a_z - r_{(4,8)z})*sin(\\beta_{(4,8)})*cos(\\phi_{(4,8)}),                                                                             (a_x - r_{(4,8)x})/sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2)) + (a_z - r_{(4,8)z})*cos(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[0, 0, 0, 0, 0, 0, 0, 0, (a_x - r_{(4,5)x})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}) - (a_y - r_{(4,5)y})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                   -cos(\\beta_{(4,5)}), -(a_x - r_{(4,8)x})*sin(\\beta_{(4,8)})*sin(\\phi_{(4,8)}) + (a_y - r_{(4,8)y})*sin(\\beta_{(4,8)})*cos(\\phi_{(4,8)}), (a_x - r_{(4,8)x})*sin(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)})) - (a_y - r_{(4,8)y})*cos(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[0, 0, 0, 0, 0, 0, 0, 0,                                                                             -sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                                     0,                                                                               sin(\\beta_{(4,8)})*cos(\\phi_{(4,8)}),                                                                                                                                                                                              -cos(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[0, 0, 0, 0, 0, 0, 0, 0,                                                                             -sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                                     0,                                                                               sin(\\beta_{(4,8)})*sin(\\phi_{(4,8)}),                                                                                                                                                                                              -sin(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[0, 0, 0, 0, 0, 0, 0, 0,                                                                                               -cos(\\beta_{(4,5)}),                                     0,                                                                                                 cos(\\beta_{(4,8)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))]]), Matrix([\n[-sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}), sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                                    0],\n[-sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}), sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                                    0],\n[                  -cos(\\beta_{(1,5)}),                   cos(\\beta_{(1,6)}),                                    0],\n[-sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}),                                    0, sin(\\beta_{(1,7)})*cos(\\phi_{(1,7)})],\n[-sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}),                                    0, sin(\\beta_{(1,7)})*sin(\\phi_{(1,7)})],\n[                  -cos(\\beta_{(1,5)}),                                    0,                   cos(\\beta_{(1,7)})],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0]]), Matrix([\n[                                                                                                                0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                   (a_y - r_{(3,6)y})*cos(\\beta_{(3,6)}) - (a_z - r_{(3,6)z})*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}), -sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}),                   -(a_y - r_{(3,8)y})*cos(\\beta_{(3,8)}) + (a_z - r_{(3,8)z})*sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}),                                                                            -(a_y - r_{(3,8)y})/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)) - (a_z - r_{(3,8)z})*sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                   -(a_y - r_{(4,5)y})*cos(\\beta_{(4,5)}) + (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}), sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                   (a_y - r_{(4,8)y})*cos(\\beta_{(4,8)}) - (a_z - r_{(4,8)z})*sin(\\beta_{(4,8)})*sin(\\phi_{(4,8)}),                                                                              (a_y - r_{(4,8)y})/sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2)) + (a_z - r_{(4,8)z})*sin(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[                                                                                                                0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                  -(a_x - r_{(3,6)x})*cos(\\beta_{(3,6)}) + (a_z - r_{(3,6)z})*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}), -sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}),                    (a_x - r_{(3,8)x})*cos(\\beta_{(3,8)}) - (a_z - r_{(3,8)z})*sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}),                                                                             (a_x - r_{(3,8)x})/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)) + (a_z - r_{(3,8)z})*cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                    (a_x - r_{(4,5)x})*cos(\\beta_{(4,5)}) - (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}), sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                  -(a_x - r_{(4,8)x})*cos(\\beta_{(4,8)}) + (a_z - r_{(4,8)z})*sin(\\beta_{(4,8)})*cos(\\phi_{(4,8)}),                                                                             -(a_x - r_{(4,8)x})/sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2)) - (a_z - r_{(4,8)z})*cos(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[                                                                                                                0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0, (a_x - r_{(3,6)x})*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}) - (a_y - r_{(3,6)y})*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}),                   -cos(\\beta_{(3,6)}), -(a_x - r_{(3,8)x})*sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}) + (a_y - r_{(3,8)y})*sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}), (a_x - r_{(3,8)x})*sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})) - (a_y - r_{(3,8)y})*cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})), -(a_x - r_{(4,5)x})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}) + (a_y - r_{(4,5)y})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                   cos(\\beta_{(4,5)}), (a_x - r_{(4,8)x})*sin(\\beta_{(4,8)})*sin(\\phi_{(4,8)}) - (a_y - r_{(4,8)y})*sin(\\beta_{(4,8)})*cos(\\phi_{(4,8)}), -(a_x - r_{(4,8)x})*sin(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)})) + (a_y - r_{(4,8)y})*cos(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[                  (a_y - r_{(2,7)y})*cos(\\beta_{(2,7)}) - (a_z - r_{(2,7)z})*sin(\\beta_{(2,7)})*sin(\\phi_{(2,7)}), -sin(\\beta_{(2,7)})*cos(\\phi_{(2,7)}),                   -(a_y - r_{(2,8)y})*cos(\\beta_{(2,8)}) + (a_z - r_{(2,8)z})*sin(\\beta_{(2,8)})*sin(\\phi_{(2,8)}),                                                                            -(a_y - r_{(2,8)y})/sqrt(1 + 1/(cos(\\delta_{(2,8)} - \\phi_{(2,8)})**2*tan(\\beta_{(2,8)})**2)) - (a_z - r_{(2,8)z})*sin(\\delta_{(2,8)})/(sqrt(1 + 1/(cos(\\delta_{(2,8)} - \\phi_{(2,8)})**2*tan(\\beta_{(2,8)})**2))*cos(\\delta_{(2,8)} - \\phi_{(2,8)})*tan(\\beta_{(2,8)})),                                                                                                                 0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                   -(a_y - r_{(4,5)y})*cos(\\beta_{(4,5)}) + (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}), sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                   (a_y - r_{(4,8)y})*cos(\\beta_{(4,8)}) - (a_z - r_{(4,8)z})*sin(\\beta_{(4,8)})*sin(\\phi_{(4,8)}),                                                                              (a_y - r_{(4,8)y})/sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2)) + (a_z - r_{(4,8)z})*sin(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[                 -(a_x - r_{(2,7)x})*cos(\\beta_{(2,7)}) + (a_z - r_{(2,7)z})*sin(\\beta_{(2,7)})*cos(\\phi_{(2,7)}), -sin(\\beta_{(2,7)})*sin(\\phi_{(2,7)}),                    (a_x - r_{(2,8)x})*cos(\\beta_{(2,8)}) - (a_z - r_{(2,8)z})*sin(\\beta_{(2,8)})*cos(\\phi_{(2,8)}),                                                                             (a_x - r_{(2,8)x})/sqrt(1 + 1/(cos(\\delta_{(2,8)} - \\phi_{(2,8)})**2*tan(\\beta_{(2,8)})**2)) + (a_z - r_{(2,8)z})*cos(\\delta_{(2,8)})/(sqrt(1 + 1/(cos(\\delta_{(2,8)} - \\phi_{(2,8)})**2*tan(\\beta_{(2,8)})**2))*cos(\\delta_{(2,8)} - \\phi_{(2,8)})*tan(\\beta_{(2,8)})),                                                                                                                 0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                    (a_x - r_{(4,5)x})*cos(\\beta_{(4,5)}) - (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}), sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                  -(a_x - r_{(4,8)x})*cos(\\beta_{(4,8)}) + (a_z - r_{(4,8)z})*sin(\\beta_{(4,8)})*cos(\\phi_{(4,8)}),                                                                             -(a_x - r_{(4,8)x})/sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2)) - (a_z - r_{(4,8)z})*cos(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[(a_x - r_{(2,7)x})*sin(\\beta_{(2,7)})*sin(\\phi_{(2,7)}) - (a_y - r_{(2,7)y})*sin(\\beta_{(2,7)})*cos(\\phi_{(2,7)}),                   -cos(\\beta_{(2,7)}), -(a_x - r_{(2,8)x})*sin(\\beta_{(2,8)})*sin(\\phi_{(2,8)}) + (a_y - r_{(2,8)y})*sin(\\beta_{(2,8)})*cos(\\phi_{(2,8)}), (a_x - r_{(2,8)x})*sin(\\delta_{(2,8)})/(sqrt(1 + 1/(cos(\\delta_{(2,8)} - \\phi_{(2,8)})**2*tan(\\beta_{(2,8)})**2))*cos(\\delta_{(2,8)} - \\phi_{(2,8)})*tan(\\beta_{(2,8)})) - (a_y - r_{(2,8)y})*cos(\\delta_{(2,8)})/(sqrt(1 + 1/(cos(\\delta_{(2,8)} - \\phi_{(2,8)})**2*tan(\\beta_{(2,8)})**2))*cos(\\delta_{(2,8)} - \\phi_{(2,8)})*tan(\\beta_{(2,8)})),                                                                                                                 0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0, -(a_x - r_{(4,5)x})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}) + (a_y - r_{(4,5)y})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                   cos(\\beta_{(4,5)}), (a_x - r_{(4,8)x})*sin(\\beta_{(4,8)})*sin(\\phi_{(4,8)}) - (a_y - r_{(4,8)y})*sin(\\beta_{(4,8)})*cos(\\phi_{(4,8)}), -(a_x - r_{(4,8)x})*sin(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)})) + (a_y - r_{(4,8)y})*cos(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[                                                                                                                0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                             -sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}),                                     0,                                                                               sin(\\beta_{(3,8)})*cos(\\phi_{(3,8)}),                                                                                                                                                                                              -cos(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                                                                               sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                                    0,                                                                             -sin(\\beta_{(4,8)})*cos(\\phi_{(4,8)}),                                                                                                                                                                                                cos(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[                                                                                                                0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                             -sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}),                                     0,                                                                               sin(\\beta_{(3,8)})*sin(\\phi_{(3,8)}),                                                                                                                                                                                              -sin(\\delta_{(3,8)})/(sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2))*cos(\\delta_{(3,8)} - \\phi_{(3,8)})*tan(\\beta_{(3,8)})),                                                                               sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                                    0,                                                                             -sin(\\beta_{(4,8)})*sin(\\phi_{(4,8)}),                                                                                                                                                                                                sin(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[                                                                                                                0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                               -cos(\\beta_{(3,6)}),                                     0,                                                                                                 cos(\\beta_{(3,8)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(3,8)} - \\phi_{(3,8)})**2*tan(\\beta_{(3,8)})**2)),                                                                                                 cos(\\beta_{(4,5)}),                                    0,                                                                                               -cos(\\beta_{(4,8)}),                                                                                                                                                                                                                                                                         -1/sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))],\n[                                                                            -sin(\\beta_{(2,7)})*cos(\\phi_{(2,7)}),                                     0,                                                                               sin(\\beta_{(2,8)})*cos(\\phi_{(2,8)}),                                                                                                                                                                                              -cos(\\delta_{(2,8)})/(sqrt(1 + 1/(cos(\\delta_{(2,8)} - \\phi_{(2,8)})**2*tan(\\beta_{(2,8)})**2))*cos(\\delta_{(2,8)} - \\phi_{(2,8)})*tan(\\beta_{(2,8)})),                                                                                                                 0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                               sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                                    0,                                                                             -sin(\\beta_{(4,8)})*cos(\\phi_{(4,8)}),                                                                                                                                                                                                cos(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[                                                                            -sin(\\beta_{(2,7)})*sin(\\phi_{(2,7)}),                                     0,                                                                               sin(\\beta_{(2,8)})*sin(\\phi_{(2,8)}),                                                                                                                                                                                              -sin(\\delta_{(2,8)})/(sqrt(1 + 1/(cos(\\delta_{(2,8)} - \\phi_{(2,8)})**2*tan(\\beta_{(2,8)})**2))*cos(\\delta_{(2,8)} - \\phi_{(2,8)})*tan(\\beta_{(2,8)})),                                                                                                                 0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                               sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                                    0,                                                                             -sin(\\beta_{(4,8)})*sin(\\phi_{(4,8)}),                                                                                                                                                                                                sin(\\delta_{(4,8)})/(sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))*cos(\\delta_{(4,8)} - \\phi_{(4,8)})*tan(\\beta_{(4,8)}))],\n[                                                                                              -cos(\\beta_{(2,7)}),                                     0,                                                                                                 cos(\\beta_{(2,8)}),                                                                                                                                                                                                                                                                         1/sqrt(1 + 1/(cos(\\delta_{(2,8)} - \\phi_{(2,8)})**2*tan(\\beta_{(2,8)})**2)),                                                                                                                 0,                                     0,                                                                                                                  0,                                                                                                                                                                                                                                                                                                                                                   0,                                                                                                 cos(\\beta_{(4,5)}),                                    0,                                                                                               -cos(\\beta_{(4,8)}),                                                                                                                                                                                                                                                                         -1/sqrt(1 + 1/(cos(\\delta_{(4,8)} - \\phi_{(4,8)})**2*tan(\\beta_{(4,8)})**2))]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[-7.36567729e-01,  2.57778286e-01,  1.67634541e-01],
       [-2.87655516e-01,  3.46060267e-01,  1.89303930e-02],
       [-2.98106622e-01,  8.86662480e-01,  2.96328113e-02],
       [ 7.18106998e-02,  1.40146742e-01, -8.36927526e-04],
       [-4.89137750e-01,  9.16881638e-02,  5.80937495e-02],
       [ 1.24023458e+00,  4.56132778e-01, -2.28362662e-01]])).all()


def test_spatial_for_descending_4():
    A = numpy.array(
      [[9, 0, 2, 0, 2, 0, 2, 0],
       [0, 9, 0, 0, 0, 0, 6, 4],
       [1, 0, 9, 0, 0, 6, 0, 0],
       [0, 0, 0, 9, 6, 0, 0, 4],
       [1, 0, 0, 0, 9, 0, 0, 0],
       [0, 0, 0, 0, 0, 9, 0, 4],
       [1, 0, 0, 0, 0, 0, 9, 0],
       [0, 0, 0, 0, 0, 0, 0, 9]])
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[sin(\\beta_{(1,3)})*cos(\\phi_{(1,3)}), 0, 0],\n[sin(\\beta_{(1,3)})*sin(\\phi_{(1,3)}), 0, 0],\n[                  cos(\\beta_{(1,3)}), 0, 0],\n[                                   0, 0, 0],\n[                                   0, 0, 0],\n[                                   0, 0, 0]]), Matrix([\n[0, 0, 0, 0, p_{(3,6)}*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)})/(2*pi) - (a_y - r_{(3,6)y})*cos(\\beta_{(3,6)}) + (a_z - r_{(3,6)z})*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}), 0, 0, 0, 0,                 0,  a_z - r_{(6,8)z}, -a_y + r_{(6,8)y}],\n[0, 0, 0, 0, p_{(3,6)}*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)})/(2*pi) + (a_x - r_{(3,6)x})*cos(\\beta_{(3,6)}) - (a_z - r_{(3,6)z})*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}), 0, 0, 0, 0, -a_z + r_{(6,8)z},                 0,  a_x - r_{(6,8)x}],\n[0, 0, 0, 0, p_{(3,6)}*cos(\\beta_{(3,6)})/(2*pi) - (a_x - r_{(3,6)x})*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}) + (a_y - r_{(3,6)y})*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}), 0, 0, 0, 0,  a_y - r_{(6,8)y}, -a_x + r_{(6,8)x},                 0],\n[0, 0, 0, 0,                                                                                                                    sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}), 0, 0, 0, 0,                 1,                 0,                 0],\n[0, 0, 0, 0,                                                                                                                    sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}), 0, 0, 0, 0,                 0,                 1,                 0],\n[0, 0, 0, 0,                                                                                                                                      cos(\\beta_{(3,6)}), 0, 0, 0, 0,                 0,                 0,                 1]]), Matrix([\n[-sin(\\beta_{(1,3)})*cos(\\phi_{(1,3)}), sin(\\beta_{(1,5)})*cos(\\phi_{(1,5)}),                                    0],\n[-sin(\\beta_{(1,3)})*sin(\\phi_{(1,3)}), sin(\\beta_{(1,5)})*sin(\\phi_{(1,5)}),                                    0],\n[                  -cos(\\beta_{(1,3)}),                   cos(\\beta_{(1,5)}),                                    0],\n[-sin(\\beta_{(1,3)})*cos(\\phi_{(1,3)}),                                    0, sin(\\beta_{(1,7)})*cos(\\phi_{(1,7)})],\n[-sin(\\beta_{(1,3)})*sin(\\phi_{(1,3)}),                                    0, sin(\\beta_{(1,7)})*sin(\\phi_{(1,7)})],\n[                  -cos(\\beta_{(1,3)}),                                    0,                   cos(\\beta_{(1,7)})],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0],\n[                                    0,                                    0,                                    0]]), Matrix([\n[                                                                                                                                                       0,                 0,                 0,                 0, -p_{(3,6)}*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)})/(2*pi) + (a_y - r_{(3,6)y})*cos(\\beta_{(3,6)}) - (a_z - r_{(3,6)z})*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}), -p_{(4,5)}*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)})/(2*pi) + (a_y - r_{(4,5)y})*cos(\\beta_{(4,5)}) - (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                 0,  a_z - r_{(4,8)z}, -a_y + r_{(4,8)y},                 0, -a_z + r_{(6,8)z},  a_y - r_{(6,8)y}],\n[                                                                                                                                                       0,                 0,                 0,                 0, -p_{(3,6)}*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)})/(2*pi) - (a_x - r_{(3,6)x})*cos(\\beta_{(3,6)}) + (a_z - r_{(3,6)z})*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}), -p_{(4,5)}*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)})/(2*pi) - (a_x - r_{(4,5)x})*cos(\\beta_{(4,5)}) + (a_z - r_{(4,5)z})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}), -a_z + r_{(4,8)z},                 0,  a_x - r_{(4,8)x},  a_z - r_{(6,8)z},                 0, -a_x + r_{(6,8)x}],\n[                                                                                                                                                       0,                 0,                 0,                 0, -p_{(3,6)}*cos(\\beta_{(3,6)})/(2*pi) + (a_x - r_{(3,6)x})*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}) - (a_y - r_{(3,6)y})*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}), -p_{(4,5)}*cos(\\beta_{(4,5)})/(2*pi) + (a_x - r_{(4,5)x})*sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}) - (a_y - r_{(4,5)y})*sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),  a_y - r_{(4,8)y}, -a_x + r_{(4,8)x},                 0, -a_y + r_{(6,8)y},  a_x - r_{(6,8)x},                 0],\n[-p_{(2,7)}*sin(\\beta_{(2,7)})*cos(\\phi_{(2,7)})/(2*pi) + (a_y - r_{(2,7)y})*cos(\\beta_{(2,7)}) - (a_z - r_{(2,7)z})*sin(\\beta_{(2,7)})*sin(\\phi_{(2,7)}),                 0,  a_z - r_{(2,8)z}, -a_y + r_{(2,8)y}, -p_{(3,6)}*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)})/(2*pi) + (a_y - r_{(3,6)y})*cos(\\beta_{(3,6)}) - (a_z - r_{(3,6)z})*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}),                                                                                                                                                        0,                 0,                 0,                 0,                 0, -a_z + r_{(6,8)z},  a_y - r_{(6,8)y}],\n[-p_{(2,7)}*sin(\\beta_{(2,7)})*sin(\\phi_{(2,7)})/(2*pi) - (a_x - r_{(2,7)x})*cos(\\beta_{(2,7)}) + (a_z - r_{(2,7)z})*sin(\\beta_{(2,7)})*cos(\\phi_{(2,7)}), -a_z + r_{(2,8)z},                 0,  a_x - r_{(2,8)x}, -p_{(3,6)}*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)})/(2*pi) - (a_x - r_{(3,6)x})*cos(\\beta_{(3,6)}) + (a_z - r_{(3,6)z})*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}),                                                                                                                                                        0,                 0,                 0,                 0,  a_z - r_{(6,8)z},                 0, -a_x + r_{(6,8)x}],\n[-p_{(2,7)}*cos(\\beta_{(2,7)})/(2*pi) + (a_x - r_{(2,7)x})*sin(\\beta_{(2,7)})*sin(\\phi_{(2,7)}) - (a_y - r_{(2,7)y})*sin(\\beta_{(2,7)})*cos(\\phi_{(2,7)}),  a_y - r_{(2,8)y}, -a_x + r_{(2,8)x},                 0, -p_{(3,6)}*cos(\\beta_{(3,6)})/(2*pi) + (a_x - r_{(3,6)x})*sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}) - (a_y - r_{(3,6)y})*sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}),                                                                                                                                                        0,                 0,                 0,                 0, -a_y + r_{(6,8)y},  a_x - r_{(6,8)x},                 0],\n[                                                                                                                                                       0,                 0,                 0,                 0,                                                                                                                    -sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}),                                                                                                                    -sin(\\beta_{(4,5)})*cos(\\phi_{(4,5)}),                 1,                 0,                 0,                -1,                 0,                 0],\n[                                                                                                                                                       0,                 0,                 0,                 0,                                                                                                                    -sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}),                                                                                                                    -sin(\\beta_{(4,5)})*sin(\\phi_{(4,5)}),                 0,                 1,                 0,                 0,                -1,                 0],\n[                                                                                                                                                       0,                 0,                 0,                 0,                                                                                                                                      -cos(\\beta_{(3,6)}),                                                                                                                                      -cos(\\beta_{(4,5)}),                 0,                 0,                 1,                 0,                 0,                -1],\n[                                                                                                                   -sin(\\beta_{(2,7)})*cos(\\phi_{(2,7)}),                 1,                 0,                 0,                                                                                                                    -sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}),                                                                                                                                                        0,                 0,                 0,                 0,                -1,                 0,                 0],\n[                                                                                                                   -sin(\\beta_{(2,7)})*sin(\\phi_{(2,7)}),                 0,                 1,                 0,                                                                                                                    -sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}),                                                                                                                                                        0,                 0,                 0,                 0,                 0,                -1,                 0],\n[                                                                                                                                     -cos(\\beta_{(2,7)}),                 0,                 0,                 1,                                                                                                                                      -cos(\\beta_{(3,6)}),                                                                                                                                                        0,                 0,                 0,                 0,                 0,                 0,                -1]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[ 1.13753423,  0.02717445, -0.96250695],
       [-0.26269499,  0.46037541,  0.37521035],
       [ 0.18613213,  0.26058143,  0.49280133],
       [-1.44841968, -1.52406937,  3.11925828],
       [ 1.47625434, -0.74717057, -0.62858022],
       [-1.6271083 ,  0.99012471,  1.06138785]])).all()

def test_spatial_for_descending_5():
    A = numpy.array(
      [[9, 1, 0, 0, 0, 1, 1, 0],
       [1, 9, 0, 0, 7, 0, 0, 0],
       [0, 0, 9, 0, 0, 7, 0, 4],
       [0, 0, 0, 9, 0, 0, 7, 4],
       [0, 0, 0, 0, 9, 0, 0, 4],
       [1, 0, 0, 0, 0, 9, 0, 0],
       [1, 0, 0, 0, 0, 0, 9, 0],
       [0, 0, 0, 0, 0, 0, 0, 9]])
    jac = Jacobian(A, robot_type = 'spatial')
    jac_func = jac.get_jacobian_function()
    assert str([jac.Ja, jac.Jp, jac.Aa, jac.Ap]) == '[Matrix([\n[                  -(a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), 0, 0],\n[                   (a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0],\n[-(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) + (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0],\n[                                                                              sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), 0, 0],\n[                                                                              sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}), 0, 0],\n[                                                                                                cos(\\beta_{(1,2)}), 0, 0]]), Matrix([\n[sin(\\beta_{(2,5)})*cos(\\phi_{(2,5)}), 0, 0, 0, 0, 0, 0, 0, 0,                 0,  a_z - r_{(5,8)z}, -a_y + r_{(5,8)y}],\n[sin(\\beta_{(2,5)})*sin(\\phi_{(2,5)}), 0, 0, 0, 0, 0, 0, 0, 0, -a_z + r_{(5,8)z},                 0,  a_x - r_{(5,8)x}],\n[                  cos(\\beta_{(2,5)}), 0, 0, 0, 0, 0, 0, 0, 0,  a_y - r_{(5,8)y}, -a_x + r_{(5,8)x},                 0],\n[                                   0, 0, 0, 0, 0, 0, 0, 0, 0,                 1,                 0,                 0],\n[                                   0, 0, 0, 0, 0, 0, 0, 0, 0,                 0,                 1,                 0],\n[                                   0, 0, 0, 0, 0, 0, 0, 0, 0,                 0,                 0,                 1]]), Matrix([\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                   -(a_y - r_{(1,6)y})*cos(\\beta_{(1,6)}) + (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                                                                                                                  0],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                    (a_x - r_{(1,6)x})*cos(\\beta_{(1,6)}) - (a_z - r_{(1,6)z})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                                                                                                                  0],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}), -(a_x - r_{(1,6)x})*sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}) + (a_y - r_{(1,6)y})*sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                                                                                                                  0],\n[                  (a_y - r_{(1,2)y})*cos(\\beta_{(1,2)}) - (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                  0,                   -(a_y - r_{(1,7)y})*cos(\\beta_{(1,7)}) + (a_z - r_{(1,7)z})*sin(\\beta_{(1,7)})*sin(\\phi_{(1,7)})],\n[                 -(a_x - r_{(1,2)x})*cos(\\beta_{(1,2)}) + (a_z - r_{(1,2)z})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                  0,                    (a_x - r_{(1,7)x})*cos(\\beta_{(1,7)}) - (a_z - r_{(1,7)z})*sin(\\beta_{(1,7)})*cos(\\phi_{(1,7)})],\n[(a_x - r_{(1,2)x})*sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}) - (a_y - r_{(1,2)y})*sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                  0, -(a_x - r_{(1,7)x})*sin(\\beta_{(1,7)})*sin(\\phi_{(1,7)}) + (a_y - r_{(1,7)y})*sin(\\beta_{(1,7)})*cos(\\phi_{(1,7)})],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                               sin(\\beta_{(1,6)})*cos(\\phi_{(1,6)}),                                                                                                                  0],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                               sin(\\beta_{(1,6)})*sin(\\phi_{(1,6)}),                                                                                                                  0],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                 cos(\\beta_{(1,6)}),                                                                                                                  0],\n[                                                                            -sin(\\beta_{(1,2)})*cos(\\phi_{(1,2)}),                                                                                                                  0,                                                                               sin(\\beta_{(1,7)})*cos(\\phi_{(1,7)})],\n[                                                                            -sin(\\beta_{(1,2)})*sin(\\phi_{(1,2)}),                                                                                                                  0,                                                                               sin(\\beta_{(1,7)})*sin(\\phi_{(1,7)})],\n[                                                                                              -cos(\\beta_{(1,2)}),                                                                                                                  0,                                                                                                 cos(\\beta_{(1,7)})]]), Matrix([\n[-sin(\\beta_{(2,5)})*cos(\\phi_{(2,5)}), -sin(\\beta_{(3,6)})*cos(\\phi_{(3,6)}),                 0,  a_z - r_{(3,8)z}, -a_y + r_{(3,8)y},                                     0,                 0,                 0,                 0,                 0, -a_z + r_{(5,8)z},  a_y - r_{(5,8)y}],\n[-sin(\\beta_{(2,5)})*sin(\\phi_{(2,5)}), -sin(\\beta_{(3,6)})*sin(\\phi_{(3,6)}), -a_z + r_{(3,8)z},                 0,  a_x - r_{(3,8)x},                                     0,                 0,                 0,                 0,  a_z - r_{(5,8)z},                 0, -a_x + r_{(5,8)x}],\n[                  -cos(\\beta_{(2,5)}),                   -cos(\\beta_{(3,6)}),  a_y - r_{(3,8)y}, -a_x + r_{(3,8)x},                 0,                                     0,                 0,                 0,                 0, -a_y + r_{(5,8)y},  a_x - r_{(5,8)x},                 0],\n[-sin(\\beta_{(2,5)})*cos(\\phi_{(2,5)}),                                     0,                 0,                 0,                 0, -sin(\\beta_{(4,7)})*cos(\\phi_{(4,7)}),                 0,  a_z - r_{(4,8)z}, -a_y + r_{(4,8)y},                 0, -a_z + r_{(5,8)z},  a_y - r_{(5,8)y}],\n[-sin(\\beta_{(2,5)})*sin(\\phi_{(2,5)}),                                     0,                 0,                 0,                 0, -sin(\\beta_{(4,7)})*sin(\\phi_{(4,7)}), -a_z + r_{(4,8)z},                 0,  a_x - r_{(4,8)x},  a_z - r_{(5,8)z},                 0, -a_x + r_{(5,8)x}],\n[                  -cos(\\beta_{(2,5)}),                                     0,                 0,                 0,                 0,                   -cos(\\beta_{(4,7)}),  a_y - r_{(4,8)y}, -a_x + r_{(4,8)x},                 0, -a_y + r_{(5,8)y},  a_x - r_{(5,8)x},                 0],\n[                                    0,                                     0,                 1,                 0,                 0,                                     0,                 0,                 0,                 0,                -1,                 0,                 0],\n[                                    0,                                     0,                 0,                 1,                 0,                                     0,                 0,                 0,                 0,                 0,                -1,                 0],\n[                                    0,                                     0,                 0,                 0,                 1,                                     0,                 0,                 0,                 0,                 0,                 0,                -1],\n[                                    0,                                     0,                 0,                 0,                 0,                                     0,                 1,                 0,                 0,                -1,                 0,                 0],\n[                                    0,                                     0,                 0,                 0,                 0,                                     0,                 0,                 1,                 0,                 0,                -1,                 0],\n[                                    0,                                     0,                 0,                 0,                 0,                                     0,                 0,                 0,                 1,                 0,                 0,                -1]])]'
    numpy.random.seed(0)
    end_effector_parameters = numpy.random.random(3)
    numpy.random.seed(0)
    robot_parameters = numpy.random.random(len(jac.parameters))
    assert numpy.isclose(jac_func(end_effector_parameters, robot_parameters),numpy.array([[-0.23376516,  0.48065023,  0.09000072],
       [-0.66267134,  1.33328825,  0.13971023],
       [-1.12571189,  2.30881396,  0.34708804],
       [ 2.22728084, -3.05228083, -0.37756044],
       [ 0.57282722, -0.91505179,  0.13080027],
       [ 1.68820665, -1.90319672, -0.17188131]])).all()
